#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`613`-compliant **type alias** (i.e., :obj:`typing.TypeAlias`
type hint singleton) utilities.

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar import BeartypeDecorHintPep613DeprecationWarning
from beartype._util.error.utilerrwarn import issue_warning

# ....................{ REDUCERS                           }....................
def reduce_hint_pep613(hint: object, exception_prefix: str, **kwargs) -> object:
    '''
    Reduce the passed :pep:`613`-compliant **type alias** (i.e.,
    :obj:`typing.TypeAlias` type hint singleton) to the ignorable
    :class:`object` superclass.

    This reducer effectively ignores *all* :obj:`typing.TypeAlias` type hint
    singletons, which convey *no* meaningful metadata or semantics. Frankly,
    it's unclear why :pep:`613` even exists. The CPython developer community
    felt similarly, which is why :pep:`695` type aliases deprecate :pep:`613`.

    This reducer is intentionally *not* memoized (e.g., by the
    ``callable_cached`` decorator), as reducers cannot be memoized.

    Parameters
    ----------
    hint : object
        Typed dictionary to be reduced.
    exception_prefix : str, optional
        Substring prefixing exception messages raised by this reducer.

    All remaining passed arguments are silently ignored.

    Returns
    -------
    object
        Lower-level type hint currently supported by :mod:`beartype`.

    Warns
    -----
    BeartypeDecorHintPep613DeprecationWarning
        :pep:`613`-compliant type aliases have been officially deprecated by
        :pep:`695`-compliant type aliases.
    '''

    #FIXME: Improve this advice, please. When we wrote this a literal lifetime
    #ago, we didn't realize that "typing_extensions" astonishingly provides a
    #backport of the low-level C-based "typing.TypeAliasType" type available in
    #Python >= 3.12 as a high-level pure-Python
    #"typing_extensions.TypeAliasType" class available across all Python
    #versions. It's... pretty astonishing, actually! Instantiating the
    #"typing_extensions.TypeAliasType" class allows users to *MUCH* more
    #conveniently define PEP 612-compliant type aliases in older Python
    #versions. PEP 612 *DEFINITELY* should have at least mentioned this, as this
    #is essential information that makes PEP 612 *MUCH* more amenable. The
    #syntax is a bit wonky, but that's understandable:
    #    # This Python >= 3.12-specific logic...
    #    type ListOrSet[T] = list[T] | set[T]
    #    # ...is equivalent to this version-agnostic logic:
    #    T = TypeVar("T")
    #    ListOrSet = TypeAliasType("ListOrSet", list[T] | set[T], type_params=(T,))

    # Emit a non-fatal deprecation warning.
    issue_warning(
        cls=BeartypeDecorHintPep613DeprecationWarning,
        message=(
            f'{exception_prefix}PEP 613 type hint {repr(hint)} '
            f'deprecated by PEP 695. Consider either:\n'
            f'* Requiring Python >= 3.12 and refactoring PEP 613 type aliases '
            f'into PEP 695 type aliases. Note that Python < 3.12 will hate you '
            f'for this: e.g.,\n'
            f'    # Instead of this...\n'
            f'    from typing import TypeAlias\n'
            f'    alias_name: TypeAlias = alias_value\n'
            f'\n'
            f'    # ..."just" do this. Congrats. You destroyed your codebase.\n'
            f'    type alias_name = alias_value\n'
            f'* Refactoring PEP 613 type aliases into PEP 484 '
            f'"typing.NewType"-based type aliases. Note that static '
            f'type-checkers (e.g., mypy, pyright, Pyre) will hate you for '
            f'this: e.g.,\n'
            f'    # Instead of this...\n'
            f'    from typing import TypeAlias\n'
            f'    alias_name: TypeAlias = alias_value\n'
            f'\n'
            f'    # ..."just" do this. Congrats. You destroyed your codebase.\n'
            f'    from typing import NewType\n'
            f'    alias_name = NewType("alias_name", alias_value)\n'
            f'\n'
            f'Combine the above two approaches via The Ultimate Type Alias '
            f'(TUTA), a hidden ninja technique that supports all Python '
            f'versions and static type-checkers but may cause coworker heads '
            f'to pop off like in that one Kingsman scene:\n'
            f'    # Instead of this...\n'
            f'    from typing import TypeAlias\n'
            f'    alias_name: TypeAlias = alias_value\n'
            f'\n'
            f'    # ..."just" do this. If you think this sucks, know that you are not alone.\n'
            f'    from typing import TYPE_CHECKING, NewType, TypeAlias  # <-- sus af\n'
            f'    from sys import version_info  # <-- code just got real\n'
            f'    if TYPE_CHECKING:  # <-- if static type-checking, then PEP 613\n'
            f'        alias_name: TypeAlias = alias_value  # <-- grimdark coding style\n'
            f'    elif version_info >= (3, 12):  # <-- if Python >= 3.12, then PEP 695\n'
            f'        exec("type alias_name = alias_value")  # <-- eldritch abomination\n'
            f'    else:  # <-- if Python < 3.12, then PEP 484\n'
            f'        alias_name = NewType("alias_name", alias_value)  # <-- coworker gives up here\n'
        ),
    )

    # Reduce *ALL* PEP 613 type hints to an arbitrary ignorable type hint.
    return object



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **abstract syntax tree (AST) getters** (i.e., low-level callables
acquiring various properties of various nodes in the currently visited AST).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from ast import (
    AST,
    Attribute,
    Expr,
    Name,
    dump as ast_dump,
)
from beartype.typing import Optional
from beartype._data.typing.datatyping import ListStrs

# ....................{ GETTERS                            }....................
#FIXME: Unit test us up, please.
def get_node_repr_indented(node: AST) -> str:
    '''
    Human-readable string pretty-printing the contents of the passed abstract
    syntax tree (AST), complete with readable indentation.

    Parameters
    ----------
    node : AST
        AST to be pretty-printed.

    Returns
    -------
    str
        Human-readable string pretty-printing the contents of this AST.
    '''
    assert isinstance(node, AST), f'{repr(node)} not AST.'

    # Return the pretty-printed contents of this AST.
    return ast_dump(node, indent=4)  # type: ignore[call-arg]

# ....................{ GETTERS ~ attr                     }....................
#FIXME: This getter could be significantly optimized. Rather than deferring to
#get_node_attr_basenames(), this getter could instead re-implement the subset of
#get_node_attr_basenames() relevant to retrieving *ONLY* the first basename.
#Doing so avoids the need for a passed "attr_basenames" list entirely. Although
#largely trivial, this optimization is currently largely unnecessary. Why?
#Because this getter is only called sporadically for outlier edge cases. *shrug*
def get_node_attr_basename_first(
    # Mandatory parameters.
    node: AST,

    # Optional parameters.
    attr_basenames: Optional[ListStrs] = None,
) -> Optional[str]:
    '''
    First unqualified basename prefixing the possibly fully-qualified
    ``"."``-delimited name of the passed **attribute name node** (i.e.,
    :class:`ast.Name` node *or* hierarchical nesting of one or more
    :class:`ast.Attribute` nodes terminating in a :class:`ast.Name` node) if
    this node is parsable by this getter *or* :data:`None` otherwise (i.e., if
    this node is unparsable by this getter).

    Parameters
    ----------
    node : AST
        Attribute name node to be unparsed.
    attr_basenames : Optional[ListStrs], default: None
        Existing caller-defined list to be efficiently cleared, reused, and
        returned by this getter if any *or* :data:`None` otherwise, in which
        case this getter instantiates and returns a new list.

    Returns
    -------
    str
        First unqualified basename prefixing the possibly fully-qualified
        ``"."``-delimited name of this attribute name node.

    See Also
    --------
    :func:`.get_node_attr_basenames`
        Further details.
    '''

    # List of the one or more unqualified basenames comprising the possibly
    # fully-qualified "."-delimited name of this attribute name node if this
    # node is parsable by this getter *OR* the empty list otherwise.
    node_attr_basenames = get_node_attr_basenames(
        node=node, attr_basenames=attr_basenames)

    # Return either the first item of this list if this list is non-empty *OR*
    # "None" otherwise (i.e., if this list is empty).
    return node_attr_basenames[0] if node_attr_basenames else None


#FIXME: Unit test us up, please.
def get_node_attr_basenames(
    # Mandatory parameters.
    node: AST,

    # Optional parameters.
    attr_basenames: Optional[ListStrs] = None,
) -> ListStrs:
    '''
    List of the one or more unqualified basenames comprising the possibly
    fully-qualified ``"."``-delimited name of the passed **attribute name node**
    (i.e., :class:`ast.Name` node *or* hierarchical nesting of one or more
    :class:`ast.Attribute` nodes terminating in a :class:`ast.Name` node) if
    this node is parsable by this getter *or* the empty list otherwise (i.e., if
    this node is unparsable by this getter).

    This getter recursively "unparses" (i.e., decompiles) the hierarchically
    nested contents of this node, albeit without actually employing recursion.
    This getter instead internally iterates over a list of the one or more
    unqualified basenames comprising this name, enabling this iteration to
    reconstruct this name.

    A list is required, as the AST grammar hierarchically nests zero or more
    :class:`.Attribute` nodes encapsulating this name in the *reverse* order of
    the expected nesting. Specifically, an attribute name qualified by N number
    of ``"."``-delimited substrings (where N >= 3) is encapsulated by a
    hierarchical nesting of N-1 :class:`.Attribute` nodes followed by 1
    :class:`.Name` node: e.g.,

    .. code-block:: python

       @package.module.submodule.decorator
       def muh_fun(): pass

    ...which is encapsulated by this AST:

    .. code-block:: python

       FunctionDef(
           name='muh_fun',
           args=arguments(),
           body=[
               Pass()],
           decorator_list=[
               Attribute(
                   value=Attribute(
                       value=Attribute(
                           value=Name(id='package', ctx=Load()),
                           attr='module',
                           ctx=Load()),
                       attr='submodule',
                       ctx=Load()),
                   attr='decorator',
                   ctx=Load()),
           ])

    That is, the :class:`.attr` instance variable of the *outermost*
    :class:`Attribute` node yields the *last* ``"."``-delimited substring of the
    fully-qualified name of that decorator. This reconstruction algorithm thus
    resembles Reverse Polish Notation, for those familiar with ancient
    calculators that no longer exist. So, nobody.

    Parameters
    ----------
    node : AST
        Attribute name node to be unparsed.
    attr_basenames : Optional[ListStrs], default: None
        Existing caller-defined list to be efficiently cleared, reused, and
        returned by this getter if any *or* :data:`None` otherwise, in which
        case this getter instantiates and returns a new list.

    Returns
    -------
    str
        Fully-qualified name *or* unqualified basename of this name node.
    '''
    assert isinstance(node, AST), f'{repr(node)} not AST.'

    # In theory, this algorithm could also be implemented with an equivalent
    # trivial one-liner resembling:
    #     return ast.unparse(node).split('.')
    #
    # In practice, doing so would:
    # * Be *PROHIBITIVELY* expensive. Recursively unparsing a node into a string
    #   merely to parse that string back into a list of strings constitutes
    #   recursive string-munging -- the ultimate in inefficient one-liners. In
    #   Python, if you want speed, you pay for speed.
    # * Probably fall down in pernicious edge cases in which the returned string
    #   is *NOT* reasonably splittable on all "." delimiters (e.g.,
    #   "obj_name.attr_name['this.is.gonna.fail.hard,yo!']").

    # ....................{ DEFAULTS                       }....................
    # True only if this getter internally instantiates a new list rather than
    # reusing an existing list passed by the caller.
    is_attr_basenames_new = attr_basenames is None

    # If the caller explicitly passed *NO* pre-initialized list, initialize this
    # to the empty list.
    if is_attr_basenames_new:
        attr_basenames = []
    # Else, the caller explicitly passed a pre-initialized list. In this case...
    else:
        assert isinstance(attr_basenames, list), (
            f'{repr(attr_basenames)} not list.')

        # Clear this list.
        attr_basenames.clear()
    # In either case, this local variable is now the empty list.

    # ....................{ NODE ~ expr                    }....................
    # If this is a high-level "Expr" node wrapping one or more lower-level
    # "Attribute" nodes and/or a lower-level "Name" node, unwrap this "Expr" to
    # the root node at the top of this hierarchical nesting of child nodes.
    if isinstance(node, Expr):
        node = node.value
    # Else, this is *NOT* a high-level "Expr" node. In this case, this is
    # assumed to be a lower-level "Attribute" or "Name" node.

    # ....................{ NODE ~ attribute               }....................
    # While the next unqualified basename name comprising this name is still
    # encapsulated by an "Attribute" node...
    #
    # Note that the AST grammar hierarchically nests "Attribute" nodes in the
    # *REVERSE* of the expected nesting. That is, the "attr" instance variable
    # of the *OUTERMOST* "Attribute" node yields the *LAST* "."-delimited
    # substring of the fully-qualified name of this attribute. This
    # reconstruction algorithm thus resembles Reverse Polish Notation, for those
    # familiar with ancient calculators that no longer exist. So, nobody.
    while isinstance(node, Attribute):
        # Append the unqualified basename of this parent submodule of this
        # attribute encapsulated by this "Attribute" child node to this list.
        #
        # Note that, as described above, "Attribute" child nodes are
        # hierarchically nested in the reverse of the expected order. In theory,
        # this basename should be *PREPENDED* rather than *APPENDED* to produce
        # the partially-qualified name of this decorator. In practice, doing so
        # is inefficient. Why? Because:
        # * List appending exhibits average-time O(1) constant-time complexity.
        # * List prepending exhibits average-time O(n) linear-time complexity.
        #
        # This algorithm thus prefers appending, which then necessitates this
        # list be reversed after algorithm termination. It's a small price to
        # pay for a substantial optimization.
        attr_basenames.append(node.attr)

        # Unwrap one hierarchical level of this "Attribute" parent node into its
        # next "Attribute" or "Name" child node.
        node = node.value  # type: ignore[assignment]
    # Else, this name is *NOT* encapsulated by an "Attribute" node.

    # ....................{ NODE ~ name                    }....................
    #FIXME: Also handle "ast.Subscript" nodes produced by statements resembling:
    #    muh_object.muh_var[muh_index]

    # If the trailing unqualified basename of this attribute is encapsulated by
    # a "Name" node, append this trailing unqualified basename to this list.
    if isinstance(node, Name):
        attr_basenames.append(node.id)
    # Else, the trailing unqualified basename of this attribute is *NOT*
    # encapsulated by a "Name" node. In this case...
    #
    # Note that this should *NEVER* happen. All attribute names should be
    # encapsulated by nodes handled above. However, the Python language and
    # hence AST grammar describing that language is constantly evolving. Since
    # this just happened, it is likely that a future iteration of the Python
    # language has now evolved in an unanticipated (yet, ultimately valid) way.
    # To preserve forward compatibility in @beartype with future Python
    # versions, intentionally ignore this unknown AST node type.
    #
    # Sometimes, doing nothing at all is the best thing you can do.
    else:
        # Clear this list. Returning this list in its currently incomplete state
        # would erroneously expose callers to unforeseen issues. Yet again,
        # doing nothing is preferable to doing a bad thing.
        attr_basenames.clear()

    # ....................{ RETURN                         }....................
    # Reverse this list to produce a list in the expected non-reversed order.
    #
    # If the caller explicitly passed *NO* pre-initialized list, reverse this
    # list by efficiently slicing this list in the reverse order into a new
    # list. Note this one-liner has been profiled to be slightly faster than the
    # comparable reversed() builtin. See also:
    #     https://www.geeksforgeeks.org/python/python-reversed-vs-1-which-one-is-faster
    if is_attr_basenames_new:
        attr_basenames = attr_basenames[::-1]
    # Else, the caller explicitly passed a pre-initialized list, implying the
    # caller would prefer to preserve this list rather than instantiating any
    # new list. Reverse this existing list in-place. Note this one-liner has
    # been profiled to be slightly slower than the approach pursued above.
    else:
        attr_basenames.reverse()

    # Return this non-reversed list.
    return attr_basenames



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **callable parameter length getter utilities** (i.e., low-level
functions introspecting the numbers of various kinds of parameters accepted by
arbitrary callables).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.roar._roarexc import _BeartypeUtilCallableException
from beartype.typing import Tuple
from beartype._cave._cavefast import (
    FunctionType,
    MethodBoundInstanceOrClassType,
)
from beartype._data.typing.datatyping import (
    Codeobjable,
    TypeException,
)
from beartype._util.func.utilfunccodeobj import get_func_codeobj
from inspect import (
    CO_VARARGS,
    CO_VARKEYWORDS,
)
from itertools import count

# ....................{ HINTS                              }....................
CallableArgsLens = Tuple[int, int, bool, bool]
'''
PEP-compliant type hint matching **callable parameter length metadata** returned
by the :func:`.get_func_args_lens` getter, defined as the 4-tuple
``(args_len_posonly_or_flex, args_len_kwonly, args_len_var_pos,
args_len_var_kw)`` such that:

* ``args_len_posonly_or_flex`` is the total number of all **non-keyword-only
  parameters** (i.e., both optional and mandatory positional-only, positional,
  and keyword parameters) accepted by the callable passed to that getter.
* ``args_len_kwonly`` is the total number of all **keyword-only parameters**
  (i.e., both optional and mandatory keyword-only parameters) accepted by the
  callable passed to that getter.
* ``args_len_var_pos`` is either:

  * If the callable passed to that getter accepts a variadic positional
    parameter, ``1``.
  * Else, ``0``.

* ``args_len_var_kw`` is either:

  * If the callable passed to that getter accepts a variadic keyword parameter,
    ``1``.
  * Else, ``0``.
'''

# ....................{ CONSTANTS ~ index                  }....................
# Iterator yielding the next integer incrementation starting at 0, to be safely
# deleted *AFTER* defining the following 0-based indices via this iterator.
__args_lens_index_counter = count(start=0, step=1)


ARGS_LENS_INDEX_POSONLY_OR_FLEX = next(__args_lens_index_counter)
'''
0-based index into each 4-tuple returned the :func:`.get_func_args_lens` getter
of the total number of all **non-keyword-only parameters** (i.e., both optional
and mandatory positional-only, positional, and keyword parameters) accepted by
the callable passed to that getter.
'''


ARGS_LENS_INDEX_KWONLY = next(__args_lens_index_counter)
'''
0-based index into each 4-tuple returned the :func:`.get_func_args_lens` getter
of the total number of all **keyword-only parameters** (i.e., both optional and
mandatory keyword-only parameters) accepted by the callable passed to that
getter.
'''


ARGS_LENS_INDEX_VAR_POS = next(__args_lens_index_counter)
'''
0-based index into each 4-tuple returned the :func:`.get_func_args_lens` getter
of either:

* If the callable passed to that getter accepts a variadic positional parameter,
  ``1``.
* Else, ``0``.
'''


ARGS_LENS_INDEX_VAR_KW = next(__args_lens_index_counter)
'''
0-based index into each 4-tuple returned the :func:`.get_func_args_lens` getter
of either:

* If the callable passed to that getter accepts a variadic keyword parameter,
  ``1``.
* Else, ``0``.
'''


# Delete the above counter for safety and sanity in equal measure.
del __args_lens_index_counter

# ....................{ GETTERS                            }....................
#FIXME: Unit test us up, please.
def get_func_args_len(*args, **kwargs) -> int:
    '''
    Total number of parameters accepted by the passed pure-Python callable,
    regardless of the kind of those parameters.

    Parameters
    ----------
    All parameters are passed as is to the lower-level
    :func:`.get_func_codeobj` getter.

    Returns
    -------
    int
        Total number of parameters accepted by that callable.

    Raises
    ------
    exception_cls
         If that callable is *not* pure-Python.
    '''

    # Number of various kinds of parameters accepted by that callable.
    (
        # Number of both optional and mandatory non-keyword-only parameters
        # (i.e., positional-only *AND* flexible (i.e., positional or keyword)
        # parameters) accepted by that callable.
        args_len_posonly_or_flex,
        # Number of both optional and mandatory keyword-only parameters accepted
        # by that callable.
        args_len_kwonly,
        # 1 only if that callable accepts variadic positional or keyword
        # parameters and 0 otherwise.
        is_arg_var_pos,
        is_arg_var_kw,
    ) = get_func_args_lens(*args, **kwargs)

    # Return the total number of parameters accepted by that callable.
    return (
        args_len_posonly_or_flex +
        args_len_kwonly +
        is_arg_var_pos +
        is_arg_var_kw
    )


#FIXME: [SPEED] Attempt to restore caching. See internal commentary. *sigh*
#FIXME: Unit test us up, please.
def get_func_args_lens(
    # Mandatory parameters.
    func: Codeobjable,

    # Optional parameters.
    is_unwrap: bool = True,
    exception_cls: TypeException = _BeartypeUtilCallableException,
    exception_prefix: str = '',
) -> CallableArgsLens:
    '''
    4-tuple ``(args_len_posonly_or_flex, args_len_kwonly, args_len_var_pos,
    args_len_var_kw)`` of all **callable parameter length metadata** describing
    the total number of various kinds of parameters accepted by the passed
    pure-Python callable.

    This getter is memoized (i.e., cached) via a :mod:`beartype`-specific
    ``__beartype_args_lens`` instance variable on that callable for efficiency.

    Parameters
    ----------
    func : Codeobjable
        Pure-Python callable, frame, or code object to be inspected.
    is_unwrap: bool, optional
        :data:`True` only if this getter implicitly calls the
        :func:`beartype._util.func.utilfuncwrap.unwrap_func_all` function.
        Defaults to :data:`True` for safety. See :func:`.get_func_codeobj` for
        further commentary.
    exception_cls : type, optional
        Type of exception to be raised in the event of a fatal error. Defaults
        to :class:`._BeartypeUtilCallableException`.
    exception_prefix : str, optional
        Human-readable label prefixing the message of any exception raised in
        the event of a fatal error. Defaults to the empty string.

    Returns
    -------
    CallableArgsLens
        4-tuple ``(args_len_posonly_or_flex, args_len_kwonly, args_len_var_pos,
        args_len_var_kw)`` of all **callable parameter length metadata.** See
        :data:`.CallableArgsLens` for further details.

    Raises
    ------
    exception_cls
         If that callable is *not* pure-Python.
    '''

    # Avoid circular import dependencies.
    from beartype._util.func.utilfuncwrap import unwrap_func_all_isomorphic

    # If unwrapping that callable, do so *BEFORE* obtaining the code object of
    # that callable for safety (to avoid desynchronization between the two).
    if is_unwrap:
        func = unwrap_func_all_isomorphic(func)  # type: ignore[arg-type]
    # Else, that callable is assumed to have already been unwrapped by the
    # caller. We should probably assert that, but doing so requires an
    # expensive call to hasattr(). What you gonna do?

    # Beartype-specific callable parameter length metadata previously computed
    # for this callable if any *OR* "None" otherwise.
    func_args_lens = getattr(func, '__beartype_args_lens', None)

    # If this metadata has been previously computed, return this metadata.
    if func_args_lens:
        return func_args_lens
    # Else, this metadata has *NOT* been previously computed.

    # Code object underlying the passed pure-Python callable unwrapped.
    func_codeobj = get_func_codeobj(
        func=func,
        is_unwrap=False,  # <-- "func" was already unwrapped above. I sigh.
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )

    # Bit field of OR-ed binary flags describing this callable.
    func_codeobj_flags = func_codeobj.co_flags

    # Callable parameter length metadata, defined as the 4-tuple of...
    func_args_lens = (
        # Number of both optional and mandatory non-keyword-only parameters
        # (i.e., positional-only *AND* flexible (i.e., positional or keyword)
        # parameters) accepted by that callable.
        func_codeobj.co_argcount,
        # Number of both optional and mandatory keyword-only parameters accepted
        # by that callable.
        func_codeobj.co_kwonlyargcount,
        # 1 only if that callable accepts variadic positional or keyword
        # parameters. Note that:
        # * Python implements booleans as integers. The standard "bool" type is
        #   a subclass of the standard "int" type. Notably, boolean:
        #   * "True" is literally integer "1".
        #   * "False" is literally integer "0".
        # * Coercing:
        #   * Any positive integer to a boolean is the most efficient means of
        #     reducing that integer to "1".
        #   * The integer "0" to a boolean silently reduces to a noop.
        # * If that callable accepts:
        #   * A variadic positional parameter, then the operation
        #     "func_codeobj_flags & CO_VARARGS == 4". Coercing that to a boolean
        #     then reduces that to the integer "1" as desired.
        #   * A variadic keyword parameter, then the operation
        #     "func_codeobj_flags & CO_VARKEYWORDS == 8". Coercing that to a
        #     boolean then reduces that to the integer "1" as desired.
        #
        # Black magic. Black magic is what we're trying to say.
        bool(func_codeobj_flags & CO_VARARGS),
        bool(func_codeobj_flags & CO_VARKEYWORDS),
    )

    #FIXME: Unsafe, fascinatingly. Uncommenting this induces test failure.
    # If that callable is actually pure-Python, cache this metadata onto that
    # callable as this beartype-specific instance variable.
    #
    # Note that doing so is unsafe if that callable is *NOT* pure-Python (e.g.,
    # is a C-based code object, call stack frame, or other low-level non-Python
    # "CodeObjable"). Why? Because only pure-Python callables can be
    # monkey-patched. Other "CodeObjable" objects hate monkey-patching: e.g.,
    #     AttributeError: 'code' object has no attribute '__beartype_args_lens'
    #     and no __dict__ for setting new attributes
    if isinstance(func, FunctionType):
        func.__beartype_args_lens = func_args_lens  # type: ignore[attr-defined]
    # Else, that callable is *NOT* pure-Python. In this case, avoid attempting
    # to cache this metadata anywhere.

    # Return this metadata.
    return func_args_lens

# ....................{ GETTERS ~ kind                     }....................
def get_func_args_flexible_len(
    # Mandatory parameters.
    func: Codeobjable,

    # Optional parameters.
    is_unwrap: bool = True,
    exception_cls: TypeException = _BeartypeUtilCallableException,
    exception_prefix: str = '',
) -> int:
    '''
    Number of **flexible parameters** (i.e., parameters passable as either
    positional or keyword arguments but *not* positional-only, keyword-only,
    variadic, or other more constrained kinds of parameters) accepted by the
    passed pure-Python callable.

    This getter transparently handles all of the following:

    * Conventional pure-Python callables.
    * If ``is_unwrap`` is :data:`True`:

      * Pure-Python **partials** (i.e., pure-Python callable
        :class:`functools.partial` objects directly wrapping pure-Python
        callables). If a partial is passed, this getter transparently returns
        the total number of flexible parameters accepted by the lower-level
        callable wrapped by this partial minus the number of flexible parameters
        partialized away by this partial.

    Parameters
    ----------
    func : Codeobjable
        Pure-Python callable, frame, or code object to be inspected.
    is_unwrap: bool, optional
        :data:`True` only if this getter implicitly calls the
        :func:`beartype._util.func.utilfuncwrap.unwrap_func_all` function.
        Defaults to :data:`True` for safety. See :func:`.get_func_codeobj` for
        further commentary.
    exception_cls : type, optional
        Type of exception to be raised in the event of a fatal error. Defaults
        to :class:`._BeartypeUtilCallableException`.
    exception_prefix : str, optional
        Human-readable label prefixing the message of any exception raised in
        the event of a fatal error. Defaults to the empty string.

    Returns
    -------
    int
        Number of flexible parameters accepted by this callable.

    Raises
    ------
    exception_cls
         If that callable is *not* pure-Python.
    '''

    # Avoid circular import dependencies.
    from beartype._util.api.standard.utilfunctools import (
        get_func_functools_partial_args_flexible_len,
        is_func_functools_partial,
    )
    from beartype._util.func.utilfunctest import is_func_codeobjable

    # If the passed callable is pure-Python...
    if is_func_codeobjable(func):
        # Number of various kinds of parameters accepted by that callable.
        func_args_lens = get_func_args_lens(
            func=func,
            is_unwrap=is_unwrap,
            exception_cls=exception_cls,
            exception_prefix=exception_prefix,
        )

        # Number of all non-keyword-only parameters (i.e., both optional and
        # mandatory positional-only, positional, and keyword parameters)
        # accepted by that callable.
        return func_args_lens[ARGS_LENS_INDEX_POSONLY_OR_FLEX]
    # Else, that callable has *NO* code object.
    #
    # If that callable is *NOT* actually callable, raise an exception.
    elif not callable(func):
        raise exception_cls(f'{exception_prefix}{repr(func)} uncallable.')
    # Else, that callable is callable.
    #
    # If unwrapping that callable *AND* that callable is a partial (i.e.,
    # "functools.partial" object wrapping a lower-level callable), return the
    # total number of flexible parameters accepted by the pure-Python wrappee
    # callable wrapped by this partial minus the number of flexible parameters
    # passed by this partial to this wrappee.
    elif is_unwrap and is_func_functools_partial(func):
        return get_func_functools_partial_args_flexible_len(
            func=func,
            is_unwrap=is_unwrap,
            exception_cls=exception_cls,
            exception_prefix=exception_prefix,
        )
    # Else, that callable is *NOT* a partial.
    #
    # By process of elimination, that callable *MUST* be an otherwise uncallable
    # object whose class has intentionally made that object callable by defining
    # the __call__() dunder method. Fallback to introspecting that method.

    # "__call__" attribute of that callable if any *OR* "None" otherwise (i.e.,
    # if that callable is actually uncallable).
    func_call_attr = getattr(func, '__call__', None)

    # If that callable fails to define the "__call__" attribute, that callable
    # is actually uncallable. But the callable() builtin claimed that callable
    # to be callable above. In this case, raise an exception.
    #
    # Note that this should *NEVER* happen. Nonetheless, this just happened.
    if func_call_attr is None:  # pragma: no cover
        raise exception_cls(
            f'{exception_prefix}{repr(func)} uncallable '
            f'(i.e., defines no __call__() dunder method).'
        )
    # Else, that callable defines the __call__() dunder method.

    # Return the total number of flexible parameters accepted by the pure-Python
    # wrappee callable wrapped by this bound method descriptor minus one to
    # account for the first "self" parameter implicitly
    # passed by this descriptor to that callable.
    return _get_func_boundmethod_args_flexible_len(
        func=func_call_attr,
        is_unwrap=is_unwrap,
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )


#FIXME: Unit test us up, please.
def get_func_args_nonvariadic_len(*args, **kwargs) -> int:
    '''
    Number of **non-variadic parameters** (i.e., parameters passable as either
    positional, positional-only, keyword, or keyword-only arguments) accepted by
    the passed pure-Python callable.

    Parameters
    ----------
    All parameters are passed as is to the lower-level
    :func:`.get_func_codeobj` getter.

    Returns
    -------
    int
        Number of flexible parameters accepted by this callable.

    Raises
    ------
    exception_cls
         If that callable is *not* pure-Python.
    '''

    # Number of various kinds of parameters accepted by the passed callable.
    func_args_lens = get_func_args_lens(*args, **kwargs)

    # Return the number of non-variadic parameters accepted by this callable.
    return (
        func_args_lens[ARGS_LENS_INDEX_POSONLY_OR_FLEX] +
        func_args_lens[ARGS_LENS_INDEX_KWONLY]
    )

# ....................{ PRIVATE ~ getters : len            }....................
def _get_func_boundmethod_args_flexible_len(
    # Mandatory parameters.
    func: MethodBoundInstanceOrClassType,

    # Optional parameters.
    is_unwrap: bool = True,
    exception_cls: TypeException = _BeartypeUtilCallableException,
    exception_prefix: str = '',
) -> int:
    '''
    Number of **flexible parameters** (i.e., parameters passable as either
    positional or keyword arguments but *not* positional-only, keyword-only,
    variadic, or other more constrained kinds of parameters) accepted by the
    passed **C-based bound instance method descriptor** (i.e., callable
    implicitly instantiated and assigned on the instantiation of an object whose
    class declares an instance function (whose first parameter is typically
    named ``self``)).

    Specifically, this getter transparently returns one less than the total
    number of flexible parameters accepted by the lower-level callable wrapped
    by this descriptor to account for the first ``self`` parameter implicitly
    passed by this descriptor to that callable.

    Parameters
    ----------
    func : MethodBoundInstanceOrClassType
        Bound method descriptor to be inspected.
    is_unwrap: bool, optional
        :data:`True` only if this getter implicitly calls the
        :func:`beartype._util.func.utilfuncwrap.unwrap_func_all` function.
        Defaults to :data:`True` for safety. See :func:`.get_func_codeobj` for
        further commentary.
    exception_cls : type, optional
        Type of exception to be raised in the event of a fatal error. Defaults
        to :class:`._BeartypeUtilCallableException`.
    exception_prefix : str, optional
        Human-readable label prefixing the message of any exception raised in
        the event of a fatal error. Defaults to the empty string.

    Returns
    -------
    int
        Number of flexible parameters accepted by this callable.

    Raises
    ------
    exception_cls
         If that callable is *not* pure-Python.
    '''

    # Avoid circular import dependencies.
    from beartype._util.func.utilfuncwrap import unwrap_func_boundmethod_once

    # Unbound pure-Python function encapsulated by this C-based bound method
    # descriptor bound to some callable object.
    wrappee = unwrap_func_boundmethod_once(
        func=func,
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )

    # Number of flexible parameters accepted by that function.
    #
    # Note that this recursive function call is guaranteed to immediately bottom
    # out and thus be safe for similar reasons as given above.
    wrappee_args_flexible_len = get_func_args_flexible_len(
        func=wrappee,
        is_unwrap=is_unwrap,
        exception_cls=exception_cls,
        exception_prefix=exception_prefix,
    )

    # If this number is zero, the caller maliciously defined a non-static
    # function accepting *NO* parameters. Since this paradoxically includes the
    # mandatory first "self" parameter for a bound method descriptor, it is
    # infeasible for this edge case to occur. Nonetheless, raise an exception.
    if not wrappee_args_flexible_len:  # pragma: no cover
        raise exception_cls(
            f'{exception_prefix}{repr(func)} accepts no '
            f'parameters despite being a bound instance method descriptor.'
        )
    # Else, this number is positive.

    # Return this number minus one to account for the fact that this bound
    # method descriptor implicitly passes the instance object to which this
    # method descriptor is bound as the first parameter to all calls of this
    # method descriptor.
    return wrappee_args_flexible_len - 1



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Root test configuration** (i.e., early-time configuration guaranteed to be
run by :mod:`pytest` *before* passed command-line arguments are parsed) for
this test suite.

Caveats
-------
For safety, this configuration should contain *only* early-time hooks absolutely
required by :mod:`pytest` design to be defined in this configuration. Hooks for
which this is the case (e.g., :func:`pytest_addoption`) are explicitly annotated
as such in official :mod:`pytest` documentation with a note resembling:

    Note

    This function should be implemented only in plugins or ``conftest.py`` files
    situated at the tests root directory due to how pytest discovers plugins
    during startup.

This file is the aforementioned ``conftest.py`` file "...situated at the tests
root directory."
'''

# ....................{ IMPORTS                            }....................
from collections.abc import Callable
from pytest import Function
from typing import Optional

# ....................{ HOOKS ~ ini                        }....................
def pytest_configure(config: 'pytest.Config') -> None:
    '''
    Hook programmatically configuring both the ``pytest`` comand-line interface
    (CLI) *and* top-level ``"pytest.ini"`` file.
    '''

    # Programmatically add our custom "run_in_subprocess" mark, enabling tests
    # to notify the pytest_pyfunc_call() hook that they require isolation to a
    # Python subprocess of the current Python process.
    config.addinivalue_line(
        'markers',
        f'{_MARK_NAME_SUBPROCESS}: mark test to run in an isolated subprocess',
    )

# ....................{ HOOKS ~ test : run                 }....................
def pytest_pyfunc_call(pyfuncitem: Function) -> Optional[bool]:
    '''
    Hook intercepting the call to run the passed :mod:`pytest` test function.

    Specifically, this test:

    * If this test has been decorated by our custom
      ``@pytest.mark.run_in_subprocess`` marker, runs this test in a Python
      subprocess of the current Python process isolated to this test.
    * Else, runs this test in the current Python process by deferring to the
      standard :mod:`pytest` logic for running this test.

    Parameters
    ----------
    pyfuncitem: Function
        :mod:`pytest`-specific object encapsulating the current test function
        being run.

    Returns
    -------
    Optional[bool]
        Either:

        * If this hook ran this test, :data:`True`.
        * If this hook did *not* run this test, :data:`None`.

    See Also
    --------
    https://github.com/ansible/pytest-mp/issues/15#issuecomment-1342682418
        GitHub comment by @pelson (Phil Elson) strongly inspiring this hook.
    '''

    # If this test has been decorated by our custom
    # @pytest.mark.run_in_subprocess marker...
    if _MARK_NAME_SUBPROCESS in pyfuncitem.keywords:
        # Defer hook-specific imports.
        from functools import partial
        from multiprocessing import Process
        from pytest import fail

        # Test function to be run.
        test_func = pyfuncitem.function

        # Partial object encapsulating the _run_test_in_subprocess() helper
        # bound to the current "pyfuncitem" object encapsulating this test.
        #
        # Note that this logic is robust under POSIX-compliant platforms (e.g.,
        # Linux, macOS) but *EXTREMELY* fragile under Windows, where the
        # "multiprocessing.Process" class sadly leverages the mostly unusable
        # "pickle" module rather than the mostly usable third-party "dill"
        # module. Notably, we intentionally:
        # * Leverage a "partial" object (which "pickle" silently supports)
        #   rather than a closure (which "pickle" rejects with
        #   non-human-readable exceptions).
        # * Pass the actual low-level test function being tested (which "pickle"
        #   silently supports) rather than the high-level "pyfuncitem" object
        #   encapsulating that function (which "pickle" rejects with
        #   non-human-readable exceptions).
        test_func_in_subprocess = partial(_test_func_in_subprocess, test_func)

        # Python subprocess tasked with running this test.
        test_subprocess = Process(target=test_func_in_subprocess)

        # Begin running this test in this subprocess.
        test_subprocess.start()

        # Block this parent Python process until this test completes.
        test_subprocess.join()

        # If this subprocess reports non-zero exit status, this test failed. In
        # this case...
        if test_subprocess.exitcode != 0:
            # Human-readable exception message to be raised.
            exception_message = (
                f'Test "{pyfuncitem.name}" failed in isolated subprocess with:')

            # Raise a pytest-compliant exception.
            raise fail(exception_message, pytrace=False)
        # Else, this subprocess reports zero exit status. In this case, this
        # test succeeded.

        # Notify pytest that this hook successfully ran this test.
        return True

    # Notify pytest that this hook avoided attempting to run this test, in which
    # case pytest will continue to look for a suitable runner for this test.
    return None

# ....................{ PRIVATE ~ globals                  }....................
_MARK_NAME_SUBPROCESS = 'run_in_subprocess'
'''
**Subprocess mark** (i.e., name of our custom :mod:`pytest` mark, enabling tests
to notify the :func:`.pytest_pyfunc_call` hook that they require isolation to a
Python subprocess of the current Python process).
'''

# ....................{ PRIVATE ~ classes                  }....................
class _UnbufferedOutputStream(object):
    '''
    **Unbuffered standard output stream** (i.e., proxy object encapsulating a
    buffered standard output stream by forcefully flushing that stream on all
    writes to that stream).

    See Also
    --------
    https://github.com/ansible/pytest-mp/issues/15#issuecomment-1342682418
        GitHub comment by @pelson (Phil Elson) strongly inspiring this class.
    '''

    def __init__(self, stream) -> None:
        self.stream = stream

    def write(self, data) -> None:
        self.stream.write(data)
        self.stream.flush()

    def writelines(self, datas) -> None:
        self.stream.writelines(datas)
        self.stream.flush()

    def __getattr__(self, attr: str) -> object:
        return getattr(self.stream, attr)

# ....................{ PRIVATE ~ functions                }....................
def _test_func_in_subprocess(test_func: Callable) -> object:
    '''
    Run the passed :mod:`pytest` test function isolated to a Python subprocess
    of the current Python process.

    Parameters
    ----------
    test_func : Callable
        Test function to be run.

    Returns
    -------
    object
        Arbitrary object returned by this test if any *or* :data:`None`.
    '''

    # Defer subpracess-specific imports.
    import sys

    # Monkey-patch the unbuffered standard error and output streams of this
    # subprocess with buffered equivalents, ensuring that pytest will reliably
    # capture *all* standard error and output emitted by running this test.
    sys.stderr = _UnbufferedOutputStream(sys.stderr)
    sys.stdout = _UnbufferedOutputStream(sys.stdout)

    # Run this test and return the result of doing so.
    return test_func()



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype import hook Python enhancement proposal (PEP) subpackage
initialization submodule** (i.e., data module mimicking real-world usage of
various :func:`beartype.claw` import hooks on various PEPs).
'''

# ....................{ IMPORTS                            }....................
from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_12
from beartype_test.a00_unit.data.claw.intraprocess.hookable_package.pep import (
    data_claw_pep557,
    pep526,
)

# If the active Python interpreter targets Python >= 3.12 and thus supports PEP
# 695...
if IS_PYTHON_AT_LEAST_3_12:
    # Defer version-specific imports.
    from beartype_test.a00_unit.data.claw.intraprocess.hookable_package.pep import (
        data_claw_pep695)
# Else, this interpreter targets Python < 3.12 and thus fails to support PEP
# 695.



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **path removers** (i.e., low-level callables permanently removing
on-disk files and directories in various reasonably safe and portable ways).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
# from beartype.roar._roarexc import _BeartypeUtilPathException
from beartype._data.typing.datatyping import (
    PathnameLike,
    PathnameLikeTuple,
)
from importlib.machinery import BYTECODE_SUFFIXES
from pathlib import Path

# ....................{ REMOVERS                           }....................
#FIXME: Unit test us up, please.
def remove_package_bytecode_files(package_dirname: PathnameLike) -> None:
    '''
    Permanently, silently, and recursively remove all **bytecode files** (i.e.,
    pure-Python bytecode compiled to platform-dependent temporary files residing
    in temporary ``__pycache__/`` subdirectories) of both the passed package and
    all subpackages of that package regardless of nesting depth.

    Usage
    -----
    This function is typically intended for usage in our test suite. Unit tests
    exercising :mod:`beartype` functionality that dynamically modifies the
    contents of bytecode files guarantee idempotency (i.e., reproducibility) by
    calling this function *before* exercising that functionality. Examples
    include :mod:`beartype.claw` import hooks that dynamically transform the
    abstract syntax trees (ASTs) of sample modules embedded in our test suite
    *before* permanently serializing (i.e., saving, writing) those changes back
    to disk within those bytecode files. Preventing desynchronization between
    the frequently changing implementations of those import hooks and those
    bytecode files requires calling this function beforehand.

    Caveats
    -------
    **This function is subject to subtle race conditions if multiple threads
    and/or processes concurrently attempt to mutate this package on the local
    filesystem.** Since *all* filesystem-centric logic suffers similar issues,
    we leave this issue as an exercise for the caller.

    Parameters
    ----------
    package_dirname : PathnameLike
        Absolute dirname of the package to remove all previously compiled
        bytecode files from.
    '''
    assert isinstance(package_dirname, PathnameLikeTuple), (
        f'{repr(package_dirname)} neither string nor "Path" object.')

    # Avoid circular import dependencies.
    from beartype._util.path.utilpathtest import die_unless_dir

    # High-level "Path" object encapsulating this dirname.
    package_dir = Path(package_dirname)

    # If this directory does *NOT* exist, raise an exception.
    die_unless_dir(package_dir)
    # Else, this directory exists.

    # For the "."-prefixed filetype of each type of platform-dependent bytecode
    # file generated by the current platform...
    #
    # Note that Python-specific glob syntax does *not* support disjunction
    # (i.e., alternation). In particular, POSIX-compliant glob disjunction
    # syntax "{match1,...,matchN}" is unsupported. If supported, that syntax
    # would enable this inefficient O(n) iteration to be trivially optimized
    # into a single O(1) call to the remove_paths_globbed() function.
    for BYTECODE_SUFFIX in BYTECODE_SUFFIXES:
        # Permanently and silently remove *ALL* bytecode files previously
        # compiled by Python into this "__pycache__/" subdirectory.
        remove_paths_globbed(
            dirname=package_dir,
            # Note that this filetype is already prefixed by ".". *sigh*
            glob=f'**/__pycache__/*{BYTECODE_SUFFIX}',
        )


#FIXME: Unit test us up, please.
def remove_paths_globbed(dirname: PathnameLike, glob: str) -> None:
    '''
    Permanently, silently, and possibly recursively remove *all* target files
    and empty directories from the source directory with the passed dirname
    matching the passed Python-specific glob expression.

    Note that Python-specific glob syntax is exactly that supported by the
    standard :mod:`fnmatch` module *plus* the recursive glob syntax ``"**/"``.
    Specifically, Python-specific glob syntax supports *only* the following
    small subset of POSIX-compliant glob syntax:

    * ``"*"`` matches everything.
    * ``"?"`` matches any single character.
    * ``"[seq]"`` matches any character in the substring ``"seq"``.
    * ``"[!seq]"`` matches any character not in the substring ``"seq"``.
    * ``"**/"`` matches *all* subdirectories recursively regardless of depth
      (e.g., ``"**/*.jpg"``, recursively removing all JPEG-formatted images from
      both this directory and all subdirectories of this directory).

    Caveats
    -------
    **This function silently ignores all non-empty directories matched by this
    glob expression.** Consider an alternate approach leveraging recursive
    directory tree traversal if requiring non-empty directory removal.

    **This function is subject to subtle race conditions if multiple threads
    and/or processes concurrently attempt to mutate this source directory.**
    Since *all* filesystem-centric logic suffers similar issues, we leave this
    issue as an exercise for the caller.

    **This function is currently inefficiently implemented in a single-threaded
    manner for simplicity.** This approach is appropriate when removing a small
    number of files but inappropriate when removing a large number of files. In
    the latter case, consider an alternate approach leveraging either
    multithreading or multiprocessing. See also this `popular article`_.

    .. _popular article:
        https://superfastpython.com/multithreaded-file-deletion

    Parameters
    ----------
    dirname : PathnameLike
        Dirname of the directory to remove *all* files and empty directories
        matching this glob from, specified as a **pathname-like** (i.e., either
        a low-level string possibly signifying a pathname *or* a high-level
        :class:`Path` instance definitely encapsulating a pathname).
    glob : str
        Python-specific glob expression matching *all* files and empty
        directories to be removed from this directory (e.g., ``"*.jpg"``).

    Raises
    ------
    _BeartypeUtilPathException
        If either:

        * This directory does *not* exist.
        * This directory exists but is *not* actually a directory.

    See Also
    ----------
    https://stackoverflow.com/a/38189275/2809027
        StackOverflow answer strongly inspiring this implementation.
    '''
    assert isinstance(dirname, PathnameLikeTuple), (
        f'{repr(dirname)} neither string nor "Path" object.')
    assert isinstance(glob, str), f'{repr(glob)} not string.'

    # Avoid circular import dependencies.
    from beartype._util.path.utilpathtest import die_unless_dir

    # High-level "Path" object encapsulating this dirname.
    dirname_path = Path(dirname)

    # If this directory does *NOT* exist, raise an exception.
    die_unless_dir(dirname_path)
    # Else, this directory exists.

    # For each matching pathname globbed from this dirname as a "Path" object...
    for pathname_globbed in dirname_path.glob(glob):
        # print(f'Removing globbed path "{pathname_globbed}"...')

        # If this pathname refers to a file...
        if pathname_globbed.is_file():
            #FIXME: Pass "missing_ok=True" *AFTER* dropping Python 3.7, as doing
            #so will improve the robustness of this logic against race
            #conditions.

            # Silently remove this file if feasible *OR* raise an exception.
            pathname_globbed.unlink()
        # Else, this pathname does *NOT* refer to a file.
        #
        # If this pathname refers to a (hopefully empty) subdirectory...
        elif pathname_globbed.is_dir():
            # Silently remove this empty subdirectory if feasible *OR* raise an
            # exception.
            pathname_globbed.rmdir()
        # Else, this pathname refers to neither a file *NOR* subdirectory. In
        # this case, silently ignore this pathname.



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Beartype decorator parameter code generator** (i.e., low-level callables
dynamically generating Python expressions type-checking all annotated parameters
of the callable currently being decorated by the :func:`beartype.beartype`
decorator in a general-purpose manner).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ TODO                               }....................
# All "FIXME:" comments for this submodule reside in this package's "__init__"
# submodule to improve maintainability and readability here.

# ....................{ IMPORTS                            }....................
from beartype.roar import (
    BeartypeDecorHintPepException,
    BeartypeDecorParamNameException,
)
from beartype.typing import (
    Optional,
    Set,
)
from beartype._data.code.datacodename import ARG_NAME_ARGS_NAME_KEYWORDABLE
from beartype._check.checkmake import make_code_raiser_func_pith_check
from beartype._check.convert.convmain import sanify_hint_root_func
from beartype._check.metadata.hint.hintsane import (
    HINT_SANE_IGNORABLE,
    HintSane,
)
from beartype._check.metadata.metadecor import BeartypeDecorMeta
from beartype._data.error.dataerrmagic import EXCEPTION_PLACEHOLDER
from beartype._data.func.datafuncarg import ARG_NAME_RETURN
from beartype._data.typing.datatypingport import Hint
from beartype._data.typing.datatyping import LexicalScope
from beartype._data.code.datacodefunc import (
    CODE_INIT_ARGS_LEN,
    # EXCEPTION_PREFIX_DEFAULT,
    ARG_KIND_TO_CODE_LOCALIZE,
)
from beartype._decor._nontype._wrap._wraputil import unmemoize_func_wrapper_code
from beartype._util.error.utilerrraise import reraise_exception_placeholder
from beartype._util.error.utilerrwarn import (
    # issue_warning,
    reissue_warnings_placeholder,
)
from beartype._util.func.arg.utilfuncargiter import (
    ArgKind,
    # ArgMandatory,
    iter_func_args,
)
from beartype._util.func.arg.utilfuncargtest import is_func_arg_variadic_keyword
from beartype._util.hint.utilhinttest import is_hint_needs_cls_stack
from beartype._util.kind.maplike.utilmapset import update_mapping
# from beartype._util.text.utiltextmunge import lowercase_str_char_first
from beartype._util.text.utiltextprefix import (
    prefix_callable_arg_name,
    # prefix_pith_value,
)
from beartype._data.kind.datakindiota import SENTINEL
from warnings import catch_warnings

# ....................{ CODERS                             }....................
def code_check_args(decor_meta: BeartypeDecorMeta) -> str:
    '''
    Generate a Python code snippet type-checking all annotated parameters of the
    decorated callable if any *or* the empty string otherwise (i.e., if these
    parameters are unannotated).

    Parameters
    ----------
    decor_meta : BeartypeDecorMeta
        Decorated callable to be type-checked.

    Returns
    -------
    str
        Code type-checking all annotated parameters of the decorated callable.

    Raises
    ------
    BeartypeDecorParamNameException
        If the name of any parameter declared on this callable is prefixed by
        the reserved substring ``__bear``.
    BeartypeDecorHintNonpepException
        If any type hint annotating any parameter of this callable is neither:

        * A PEP-noncompliant type hint.
        * A supported PEP-compliant type hint.
    '''
    assert isinstance(decor_meta, BeartypeDecorMeta), (
        f'{repr(decor_meta)} not beartype call.')

    # ..................{ LOCALS ~ func                      }..................
    # If *NO* callable parameters are annotated, silently reduce to a noop.
    #
    # Note that this is purely an optimization short-circuit mildly improving
    # efficiency for the common case of callables accepting either no
    # parameters *OR* one or more parameters, all of which are unannotated.
    if (
        # That callable is annotated by only one type hint *AND*...
        len(decor_meta.func_annotations) == 1 and
        # That type hint annotates that callable's return rather than a
        # parameter accepted by that callable...
        ARG_NAME_RETURN in decor_meta.func_annotations
    ):
        return ''
    # Else, one or more callable parameters are annotated.

    # Python code snippet to be returned, defaulting to the empty string
    # implying this callable's parameters to all either be unannotated *OR*
    # annotated only by safely ignorable type hints.
    func_wrapper_code = ''

    # Lexical scope (i.e., dictionary mapping from the relative unqualified name
    # to value of each locally or globally scoped attribute accessible to a
    # callable or class), initialized to "None" for safety.
    func_scope: LexicalScope = None  # type: ignore[assignment]

    # ..................{ LOCALS ~ parameter                 }..................
    #FIXME: Remove this *AFTER* optimizing signature generation, please.
    # True only if that callable possibly accepts one or more positional
    # parameters.
    is_args_positional = False

    #FIXME: ******UNIT TEST US UP, PLEASE.******* Do so exhaustively until
    #exhausted. This is super-critical. Yo!
    #FIXME: Remove the "args_name_keywordable" local variable and associated
    #"ARG_NAME_ARGS_NAME_KEYWORDABLE" global variable *AFTER* refactoring
    #@beartype to generate callable-specific wrapper signatures.

    # Either...
    args_name_keywordable: Optional[Set[str]] = (
        #FIXME: [SPEED] Minor optimization. If the decorated callable *ONLY*
        #accepts an annotated variadic keyword parameter and no other
        #parameters, then the empty set reused from a global "SET_EMPTY" import
        #suffices here. Probably. Consider, anyway. *sigh*
        # If the decorated callable accepts an annotated variadic keyword
        # parameter (e.g., "**kwargs: int"), the set of the names of all
        # keywordable parameters (i.e., parameters that may be passed by
        # keyword), defined as the union of the:
        # * Set of the names of all flexible parameters (i.e., parameters that
        #   may be passed either positionally or by keyword).
        # * Set of the names of all keyword-only parameters.
        #
        # This set is required in the body of the type-checking function
        # wrapping the decorated callable to differentiate between:
        # * Keyword parameters explicitly accepted by the decorated callable,
        #   which are trivially type-checked by existing logic.
        # * Keyword parameters *NOT* explicitly accepted by the decorated
        #   callable and thus implicitly accepted only as excess keyword
        #   parameters added by CPython itself to the annotated variadic keyword
        #   parameter (e.g., "**kwargs: int") accepted by the decorated
        #   callable. Excess keyword parameters *CANNOT* be trivially
        #   type-checked by existing logic, due to the non-triviality of
        #   deciding whether a keyword parameter even is "excess" or not.
        set()
        if is_func_arg_variadic_keyword(
            # See the call to the iter_func_args() generator function below for
            # further commentary on these parameters.
            func=decor_meta.func_wrappee_wrappee, is_unwrap=False) else
        # Else, "None".
        None
    )

    # ..................{ LOCALS ~ hint                      }..................
    # Possibly insane hint annotating the current parameter if any *OR* the
    # sentinel placeholder otherwise (i.e., if this parameter is unannotated).
    hint_insane: Hint = None  # type: ignore[assignment]

    # Sanified hint metadata annotating the current parameter, sanified from
    # this possibly insane hint.
    hint_sane: HintSane = None  # type: ignore[assignment]

    # ..................{ GENERATE                           }..................
    #FIXME: Locally remove the "arg_index" local variable (and thus avoid
    #calling the enumerate() builtin here) *AFTER* refactoring @beartype to
    #generate callable-specific wrapper signatures.

    # Kind and name of this parameter.
    arg_kind: ArgKind = None  # type: ignore[assignment]
    arg_name: str     = None  # type: ignore[assignment]

    #FIXME: Uncomment as needed, please. *sigh*
    # # Default value of this parameter if this parameter is optional *OR* the
    # # "ArgMandatory" singleton otherwise (i.e., if this parameter is mandatory).
    # arg_default: object = None

    # For the 0-based index of each parameter accepted by that callable and the
    # "ArgMeta" 3-tuple describing this parameter (in declaration order)...
    for arg_index, arg_meta in enumerate(iter_func_args(
        # Possibly lowest-level wrappee underlying the possibly higher-level
        # wrapper currently being decorated by the @beartype decorator. The
        # latter typically fails to convey the same callable metadata conveyed
        # by the former -- including the names and kinds of parameters accepted
        # by the possibly unwrapped callable. This renders the latter mostly
        # useless for our purposes.
        func=decor_meta.func_wrappee_wrappee,
        func_codeobj=decor_meta.func_wrappee_wrappee_codeobj,
        # Avoid inefficiently attempting to re-unwrap this wrappee. The
        # previously called BeartypeDecorMeta.reinit() method has already
        # guaranteed this wrappee to be isomorphically unwrapped.
        is_unwrap=False,
    )):
        # Localize metadata for both efficiency and f-string purposes.
        #
        # Note that list unpacking is substantially more efficient than
        # manually indexing list items. The former requires only a single Python
        # statement, whereas the latter requires "n" Python statements.
        (
            arg_kind,
            arg_name,
            _,  # arg_default,
        ) = arg_meta

        # If...
        if (
            # The set of the names of all keywordable parameters must be decided
            # to type-check the decorated callable *AND*...
            args_name_keywordable is not None and
            # This parameter is keywordable...
            arg_kind in _ARG_KINDS_KEYWORD
        ):
            # Add the name of this parameter to that set.
            args_name_keywordable.add(arg_name)
        # Else, this parameter *CANNOT* be passed by keyword.

        # Type hint annotating this parameter if any *OR* the sentinel
        # placeholder otherwise (i.e., if this parameter is unannotated).
        #
        # Note that "None" is a semantically meaningful PEP 484-compliant type
        # hint equivalent to "type(None)". Ergo, we *MUST* explicitly
        # distinguish between that type hint and unannotated parameters.
        hint_insane = decor_meta.func_annotations_get(  # type: ignore[assignment]
            arg_name, SENTINEL)

        # If this parameter is unannotated, continue to the next parameter.
        if hint_insane is SENTINEL:
            continue
        # Else, this parameter is annotated.

        # Attempt to...
        try:
            # With a context manager "catching" *ALL* non-fatal warnings emitted
            # during this logic for subsequent "playrback" below...
            with catch_warnings(record=True) as warnings_issued:
                # If this parameter's name is reserved for use by the @beartype
                # decorator, raise an exception.
                if arg_name.startswith('__bear'):
                    raise BeartypeDecorParamNameException(
                        f'{EXCEPTION_PLACEHOLDER}reserved by @beartype.')
                # Else, this parameter's name is *NOT* reserved for use by the
                # @beartype decorator.

                # Sane hint sanified from this possibly insane parameter hint if
                # sanifying this hint generated no supplementary metadata *OR*
                # that metadata otherwise. Additionally, if this hint is
                # unsupported by @beartype, raise an exception.
                hint_sane = sanify_hint_root_func(
                    decor_meta=decor_meta,
                    hint=hint_insane,
                    pith_name=arg_name,
                    arg_kind=arg_kind,
                    exception_prefix=EXCEPTION_PLACEHOLDER,
                )

                # If this hint is ignorable, continue to the next parameter.
                if hint_sane is HINT_SANE_IGNORABLE:
                    # print(f'Ignoring {decor_meta.func_name} parameter {arg_name} hint {repr(hint)}...')
                    continue
                # Else, this hint is unignorable.

                #FIXME: Fundamentally unsafe and thus temporarily disabled *FOR
                #THE MOMENT.* The issue is that our current implementation of
                #the is_bearable() tester internally called by this function
                #refuses to resolve relative forward references -- which is
                #obviously awful. Ideally, that tester *ABSOLUTELY* should
                #resolve relative forward references. Until it does, however,
                #this is verboten dark magic that is unsafe in the general case.
                #FIXME: Note that there exist even *MORE* edge cases, however:
                #@dataclass fields, which violate typing semantics: e.g.,
                #    from dataclasses import dataclass, field
                #    from typing import Dict
                #
                #    from beartype import beartype
                #
                #    @beartype
                #    @dataclass
                #    class A:
                #        test_dict: Dict[str, str] = field(default_factory=dict)
                #FIXME: Once this has been repaired, please reenable:
                #* The "test_decor_arg_kind_flex_optional" unit test.

                # # If this parameter is optional *AND* the default value of this
                # # optional parameter violates this hint, raise an exception.
                # _die_if_arg_default_unbearable(
                #     decor_meta=decor_meta, arg_default=arg_default, hint=hint)
                # # Else, this parameter is either optional *OR* the default value
                # # of this optional parameter satisfies this hint.

                # If this parameter either may *OR* must be passed positionally,
                # record this fact.
                #
                # Note this conditional branch *MUST* be tested after validating
                # this parameter to be unignorable; if this branch were instead
                # nested *BEFORE* validating this parameter to be unignorable,
                # beartype would fail to reduce to a noop for otherwise
                # ignorable callables -- which would be rather bad, really.
                if arg_kind in _ARG_KINDS_POSITIONAL:
                    is_args_positional = True
                # Else, this parameter *CANNOT* be passed positionally.

                #FIXME: [SPEED] Negligibly optimized the ".get" access away:
                #    ARG_LOCALIZE_TEMPLATE = ARG_KIND_TO_CODE_LOCALIZE_get(  # type: ignore
                #        arg_kind, None)
                # Python code template localizing this parameter if this kind of
                # parameter is supported *OR* "None" otherwise.
                ARG_LOCALIZE_TEMPLATE = ARG_KIND_TO_CODE_LOCALIZE.get(  # type: ignore
                    arg_kind, None)

                # If this kind of parameter is unsupported, raise an exception.
                #
                # Note this edge case should *NEVER* occur, as the parent
                # function should have simply ignored this parameter.
                if ARG_LOCALIZE_TEMPLATE is None:
                    raise BeartypeDecorHintPepException(
                        f'{EXCEPTION_PLACEHOLDER}kind {repr(arg_kind)} '
                        f'currently unsupported by @beartype.'
                    )
                # Else, this kind of parameter is supported. Ergo, this code is
                # non-"None".

                #FIXME: DRY violation. The same logic appears in "_wrapreturn"
                #as well. It looks like what we *PROBABLY* want to do here is:
                #* Rename the existing make_code_raiser_func_pith_check()
                #  factory to _make_code_raiser_func_pith_check_cached().
                #* Define a new make_code_raiser_func_pith_check() factory that
                #  is unmemoized and has a simpler public API. Notably, this new
                #  factory should:
                #  * Accept a new "decor_meta" parameter.
                #  * Drop the existing "conf" and "cls_stack" parameters.
                #  * Pass parameters by keyword rather than positionally. This
                #    would be especially useful for the "is_param" parameter,
                #    which would suddenly become readable below.
                #  * Internally compute the "cls_stack" as given below.

                # Type stack if required by this hint *OR* "None" otherwise. See
                # the is_hint_needs_cls_stack() tester for further discussion.
                #
                # Note that the original unsanitized "hint_insane" (e.g.,
                # "typing.Self") rather than the new sanitized "hint" (e.g., the
                # class currently being decorated by @beartype) is passed to
                # that tester. Why? Because the latter may already have been
                # reduced above to a different (and seemingly innocuous) type
                # hint that does *NOT* appear to require a type stack at late
                # *EXCEPTION RAISING TIME* (i.e., the
                # beartype._check.error.errmain.get_func_pith_violation()
                # function) but actually does. Only the original unsanitized
                # "hint_insane" is truth.
                cls_stack = (
                    decor_meta.cls_stack
                    if is_hint_needs_cls_stack(hint_insane) else
                    None
                )
                # print(f'arg "{arg_name}" hint {repr(hint)} cls_stack: {repr(cls_stack)}')

                # Code snippet type-checking any parameter with arbitrary name.
                (
                    code_arg_check_pith,
                    func_scope,
                    hint_refs_type_basename,
                ) = make_code_raiser_func_pith_check(
                    hint_sane,
                    decor_meta.conf,
                    cls_stack,
                    True,  # <-- True only for parameters
                )

                # Merge the local scope required to check this parameter into
                # the local scope currently required by the current wrapper
                # function.
                update_mapping(decor_meta.func_wrapper_scope, func_scope)

                # Python code snippet localizing this parameter.
                code_arg_localize = ARG_LOCALIZE_TEMPLATE.format(
                    arg_name=arg_name, arg_index=arg_index)

                # Unmemoize this snippet against the current parameter.
                code_arg_check = unmemoize_func_wrapper_code(
                    decor_meta=decor_meta,
                    func_wrapper_code=code_arg_check_pith,
                    pith_repr=repr(arg_name),
                    hint_refs_type_basename=hint_refs_type_basename,
                )

                # Append code type-checking this parameter against this hint.
                func_wrapper_code += f'{code_arg_localize}{code_arg_check}'

            # If one or more warnings were issued, reissue these warnings with
            # each placeholder substring (i.e., "EXCEPTION_PLACEHOLDER"
            # instance) replaced by a human-readable description of this
            # callable and annotated parameter.
            if warnings_issued:
                # print(f'warnings_issued: {warnings_issued}')
                reissue_warnings_placeholder(
                    warnings=warnings_issued,
                    target_str=prefix_callable_arg_name(
                        func=decor_meta.func_wrappee,
                        arg_name=arg_name,
                        is_color=decor_meta.conf.is_color,
                    ),
                )
            # Else, *NO* warnings were issued.
        # If any exception was raised, reraise this exception with each
        # placeholder substring (i.e., "EXCEPTION_PLACEHOLDER" instance)
        # replaced by a human-readable description of this callable and
        # annotated parameter.
        except Exception as exception:
            reraise_exception_placeholder(
                exception=exception,
                #FIXME: Embed the kind of parameter both here and above as well
                #(e.g., "positional-only", "keyword-only", "variadic
                #positional"), ideally by improving the existing
                #prefix_callable_arg_name() function to introspect this kind from
                #the callable code object.
                target_str=prefix_callable_arg_name(
                    func=decor_meta.func_wrappee,
                    arg_name=arg_name,
                    is_color=decor_meta.conf.is_color,
                ),
            )

    # ..................{ RETURN                             }..................
    # If that callable accepts an annotated variadic keyword parameter, expose
    # the set of the names of all keywordable parameters to this wrapper
    # function needed to type-check that annotated variadic keyword parameter.
    if args_name_keywordable is not None:
        decor_meta.func_wrapper_scope[ARG_NAME_ARGS_NAME_KEYWORDABLE] = (
            args_name_keywordable)
    # Else, that callable accepts *NO* annotated variadic parameter.

    # If that callable accepts one or more annotated positional parameters,
    # prefix this code by a snippet localizing the number of these parameters.
    if is_args_positional:
        func_wrapper_code = f'{CODE_INIT_ARGS_LEN}{func_wrapper_code}'
    # Else, that callable accepts *NO* annotated positional parameters.

    # Return this code.
    return func_wrapper_code

# ....................{ PRIVATE ~ constants                }....................
#FIXME: Shift these constants into a more appropriate "beartype._data"
#submodule, please. *sigh*
_ARG_KINDS_KEYWORD = frozenset((
    ArgKind.KEYWORD_ONLY,
    ArgKind.POSITIONAL_OR_KEYWORD,
))
'''
Frozen set of all **keyword parameter kinds** (i.e., :attr:`ArgKind` enumeration
members signifying that a callable parameter either may *or* must be passed by
keyword).
'''


_ARG_KINDS_POSITIONAL = frozenset((
    ArgKind.POSITIONAL_ONLY,
    ArgKind.POSITIONAL_OR_KEYWORD,
))
'''
Frozen set of all **positional parameter kinds** (i.e., :class:`.ArgKind`
enumeration members signifying that a callable parameter either may *or* must be
passed positionally).
'''

# ....................{ PRIVATE ~ raisers                  }....................
#FIXME: Preserved for posterity. We'll almost certainly want to restore this at
#some future date. Until then, we sigh. *sigh*
# def _die_if_arg_default_unbearable(
#     decor_meta: BeartypeDecorMeta, arg_default: object, hint: Hint) -> None:
#     '''
#     Raise a violation exception if the annotated optional parameter of the
#     decorated callable with the passed default value violates the type hint
#     annotating that parameter at decoration time.
#
#     Parameters
#     ----------
#     decor_meta : BeartypeDecorMeta
#         Decorated callable to be type-checked.
#     arg_default : object
#         Either:
#
#         * If this parameter is mandatory, the :data:`.ArgMandatory` singleton.
#         * If this parameter is optional, the default value of this optional
#           parameter to be type-checked.
#     hint : Hint
#         Type hint to type-check against this default value.
#
#     Warns
#     -----
#     BeartypeDecorHintParamDefaultForwardRefWarning
#         If this type hint contains one or more forward references that *cannot*
#         be resolved at decoration time. While this does *not* necessarily
#         constitute a fatal error from the end user perspective, this does
#         constitute a non-fatal issue worth informing the end user of.
#
#     Raises
#     ------
#     BeartypeDecorHintParamDefaultViolation
#         If this default value violates this type hint.
#     '''
#     assert isinstance(decor_meta, BeartypeDecorMeta), (
#         f'{repr(decor_meta)} not beartype call.')
#
#     # ..................{ PREAMBLE                           }..................
#     # If this parameter is mandatory, silently reduce to a noop.
#     if arg_default is ArgMandatory:
#         return
#     # Else, this parameter is optional and thus defaults to a default value.
#
#     # ..................{ IMPORTS                            }..................
#     # Defer heavyweight imports prohibited at global scope.
#     from beartype.door import (
#         die_if_unbearable,
#         is_bearable,
#     )
#
#     # ..................{ MAIN                               }..................
#     # Attempt to...
#     try:
#         # If this default value satisfies this hint, silently reduce to a noop.
#         #
#         # Note that this is a non-negligible optimization. Technically, this
#         # preliminary test is superfluous: only the call to the
#         # die_if_unbearable() raiser below is required. Pragmatically, this
#         # preliminary test avoids a needlessly expensive dictionary copy in the
#         # common case that this value satisfies this hint.
#         if is_bearable(obj=arg_default, hint=hint, conf=decor_meta.conf):
#             return
#         # Else, this default value violates this hint.
#     #FIXME: Probably generalize this to *ANY* exception whatsoever, no?
#     # If doing so raises a forward hint exception, this hint contains one or
#     # more unresolvable forward references to user-defined objects that have yet
#     # to be defined. In all likelihood, these objects are subsequently defined
#     # after the definition of this decorated callable. While this does *NOT*
#     # necessarily constitute a fatal error from the end user perspective, this
#     # does constitute a non-fatal issue worth informing the end user of. In this
#     # case, we coerce this exception into a warning.
#     except _BeartypeHintForwardRefExceptionMixin as exception:
#         # Forward hint exception message raised above. To readably embed this
#         # message in the longer warning message emitted below, the first
#         # character of this message is lowercased as well.
#         exception_message = lowercase_str_char_first(str(exception))
#
#         # Emit this non-fatal warning.
#         issue_warning(
#             cls=BeartypeDecorHintParamDefaultForwardRefWarning,
#             message=(
#                 f'{EXCEPTION_PREFIX_DEFAULT}value '
#                 f'{prefix_pith_value(pith=arg_default, is_color=decor_meta.conf.is_color)}'
#                 f'uncheckable at @beartype decoration time, as '
#                 f'{exception_message}'
#             ),
#         )
#
#         # Loudly reduce to a noop. Since this forward reference is unresolvable,
#         # further type-checking attempts are entirely fruitless.
#         return
#
#     # Modifiable keyword dictionary encapsulating this beartype configuration.
#     conf_kwargs = decor_meta.conf.kwargs.copy()
#
#     #FIXME: This should probably be configurable as well. For now, this is fine.
#     #We shrug noncommittally. We shrug, everyone! *shrug*
#     # Set the type of violation exception raised by the subsequent call to the
#     # die_if_unbearable() function to the expected type.
#     conf_kwargs['violation_door_type'] = BeartypeDecorHintParamDefaultViolation
#
#     # New beartype configuration initialized by this dictionary.
#     conf = BeartypeConf(**conf_kwargs)
#
#     # Raise this type of violation exception.
#     die_if_unbearable(
#         obj=arg_default,
#         hint=hint,
#         conf=conf,
#         exception_prefix=EXCEPTION_PREFIX_DEFAULT,
#     )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`695` **type hint utility** unit tests.

This submodule unit tests the public API of the private
:mod:`beartype._util.hint.pep.proposal.pep695` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ tester                     }....................
def test_is_hint_pep695_subbed() -> None:
    '''
    Test the private
    :mod:`beartype._util.hint.pep.proposal.pep695.is_hint_pep695_subbed`
    tester.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype._util.hint.pep.proposal.pep695 import is_hint_pep695_subbed
    from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_12

    # If the active Python interpreter targets Python >= 3.12 and thus supports
    # PEP 695...
    if IS_PYTHON_AT_LEAST_3_12:
        # Defer version-specific imports.
        from beartype_test.a00_unit.data.pep.pep695.data_pep695util import (
            unit_test_is_hint_pep695_subbed)

        # Perform this test.
        unit_test_is_hint_pep695_subbed()
    # Else, this interpreter targets Python < 3.12 and thus fails to support PEP
    # 695.

    # ....................{ FAIL                           }....................
    # Assert this tester rejects objects that are *NOT* PEP 585-compliant
    # subscripted builtins.
    assert is_hint_pep695_subbed(
        'And thou, colossal Skeleton, that, still') is False

# ....................{ TESTS ~ getter                     }....................
def test_get_hint_pep695_parameterizable_typeparams() -> None:
    '''
    Test the private
    :mod:`beartype._util.hint.pep.proposal.pep695._get_hint_pep695_parameterizable_typeparams`
    getter.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep695Exception
    from beartype._util.hint.pep.proposal.pep695 import (
        _get_hint_pep695_parameterizable_typeparams)
    from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_12
    from beartype_test.a00_unit.data.data_type import (
        Class,
        function,
    )
    from pytest import raises

    # If the active Python interpreter targets Python >= 3.12 and thus supports
    # PEP 695...
    if IS_PYTHON_AT_LEAST_3_12:
        # Defer version-specific imports.
        from beartype_test.a00_unit.data.pep.pep695.data_pep695util import (
            unit_test_get_hint_pep695_parameterizable_typeparams)

        # Perform this test.
        unit_test_get_hint_pep695_parameterizable_typeparams()
    # Else, this interpreter targets Python < 3.12 and thus fails to support PEP
    # 695.

    # Assert this getter returns the empty tuple for pure-Python classes and
    # functions that are *NOT* PEP 695-parametrized.
    assert _get_hint_pep695_parameterizable_typeparams(Class) == ()
    assert _get_hint_pep695_parameterizable_typeparams(function) == ()

    # ....................{ FAIL                           }....................
    # Assert this getter raises the expected exception when passed an arbitrary
    # object that is *NOT* parameterizable under PEP 695.
    with raises(BeartypeDecorHintPep695Exception):
        _get_hint_pep695_parameterizable_typeparams(
            'And all the gloom and sorrow of the place,')

# ....................{ TESTS ~ iterator                   }....................
def test_iter_hint_pep695_forwardrefs() -> None:
    '''
    Test the private
    :mod:`beartype._util.hint.pep.proposal.pep695.iter_hint_pep695_unsubbed_forwardrefs`
    iterator.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep695Exception
    from beartype._util.hint.pep.proposal.pep695 import (
        iter_hint_pep695_unsubbed_forwardrefs)
    from beartype._util.py.utilpyversion import IS_PYTHON_AT_LEAST_3_12
    from pytest import raises

    # If the active Python interpreter targets Python >= 3.12 and thus supports
    # PEP 695...
    if IS_PYTHON_AT_LEAST_3_12:
        # Defer version-specific imports.
        from beartype_test.a00_unit.data.pep.pep695.data_pep695util import (
            unit_test_iter_hint_pep695_forwardrefs)

        # Perform this test.
        unit_test_iter_hint_pep695_forwardrefs()
    # Else, this interpreter targets Python < 3.12 and thus fails to support PEP
    # 695.

    # ....................{ FAIL                           }....................
    # Assert this iterator raises the expected exception when passed an
    # arbitrary PEP 695-noncompliant object.
    with raises(BeartypeDecorHintPep695Exception):
        next(iter_hint_pep695_unsubbed_forwardrefs(
            'Tumultuously accorded with those fits'))



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`484`- and :pep:`585`-compliant **items view type hint
reducers** (i.e., low-level callables converting higher-level :pep:`484`- and
:pep:`585`-compliant ``ItemsView[...]` type hints to lower-level type hints more
readily consumable by :mod:`beartype`).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.typing import (
    Annotated,
    Collection,
    Tuple,
)
from beartype._data.typing.datatypingport import Hint
from beartype._util.cache.utilcachecall import callable_cached
from beartype._util.hint.pep.proposal.pep484585.pep484585 import (
    get_hint_pep484585_args)
from beartype._util.hint.pep.utilpeptest import is_hint_pep_subbed
from collections.abc import ItemsView as ItemsViewABC

# ....................{ REDUCERS                           }....................
@callable_cached
def reduce_hint_pep484585_itemsview(hint: Hint, exception_prefix: str) -> Hint:
    '''
    Reduce the passed :pep:`484`- or :pep:`585`-compliant **items view type
    hint** (i.e., of the form ``(collections.abc|typing).ItemsView[{hint_key},
    {hint_value}]``) to a more suitable type hint better supported by
    :mod:`beartype`.

    This reducer is memoized for efficiency.

    Parameters
    ----------
    hint : Hint
        Items view type hint to be reduced.
    exception_prefix : str, optional
        Human-readable substring prefixing raised exception messages.

    Returns
    -------
    Hint
        More suitable type hint better supported by :mod:`beartype`.
    '''

    # Avoid circular import dependencies.
    from beartype._check.convert._reduce._pep.pep484.redpep484 import (
        reduce_hint_pep484_deprecated)

    # If this hint is a PEP 484-compliant deprecated "typing.ItemsView[...]"
    # type hint, emit a non-fatal deprecation warning.
    reduce_hint_pep484_deprecated(hint=hint, exception_prefix=exception_prefix)

    # Reduced hint to be returned, defaulting to the abstract base class (ABC)
    # of *ALL* items views.
    hint_reduced: Hint = ItemsViewABC

    # If this hint is subscripted by one or more child type hints...
    if is_hint_pep_subbed(hint):
        # Defer heavyweight imports.
        from beartype.vale import IsInstance

        # Child key and value type hints subscripting this parent type hint.
        hint_key, hint_value = get_hint_pep484585_args(  # type: ignore[misc]
            hint=hint, args_len=2, exception_prefix=exception_prefix)

        # Reduce this hint to a PEP 593-compliant hint annotating...
        hint_reduced = Annotated[
            # A collection of 2-tuples "(key, value)" -- which, interestingly,
            # is literally what an items view is.
            #
            # Look. @beartype doesn't make the insane rules. It just enforces
            # them. We pretend this makes the world a better place.
            Collection[Tuple[hint_key, hint_value]],  # type: ignore[assignment, valid-type]
            # Constrain this collection to be an instance of the expected
            # "collections.abc.ItemsView" abstract base class (ABC).
            IsInstance[ItemsViewABC],
        ]
    # Else, this hint is unsubscripted. In this case, reduce to type-checking
    # that an items view is an instance of this ABC (via the above default).

    # Return this reduced hint.
    return hint_reduced



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Beartype **Decidedly Object-Oriented Runtime-checking (DOOR) class type hint
classes** (i.e., :class:`beartype.door.TypeHint` subclasses implementing support
for :pep:`484`-compliant type hints that are, in fact, simple classes).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
from beartype.door._cls.doorsuper import TypeHint
from beartype.typing import TYPE_CHECKING

# ....................{ SUBCLASSES                         }....................
class ClassTypeHint(TypeHint):
    '''
    **Class type hint wrapper** (i.e., high-level object encapsulating a
    low-level :pep:`484`-compliant type hint that is, in fact, a simple class).

    Caveats
    -------
    This wrapper also intentionally wraps :pep:`484`-compliant :data:``None`
    type hints as the simple type of the :data:``None` singleton, as :pep:`484`
    standardized the reduction of the former to the latter:

         When used in a type hint, the expression None is considered equivalent
         to type(None).

    Although a unique ``NoneTypeHint`` subclass of this class specific to the
    :data:`None` singleton *could* be declared, doing so is substantially
    complicated by the fact that numerous PEP-compliant type hints internally
    elide :data:`None` to the type of that singleton before the
    :mod:`beartype.door` API ever sees a distinction. Notably, this includes
    :pep:`484`-compliant unions subscripted by that singleton: e.g.,

    .. code-block:: python

       >>> from typing import Union
       >>> Union[str, None].__args__
       (str, NoneType)
    '''

    # ..................{ STATIC                             }..................
    # Squelch false negatives from static type checkers.
    if TYPE_CHECKING:
        _hint: type

    # ..................{ PRIVATE ~ properties               }..................
    @property
    def _is_args_ignorable(self) -> bool:

        # Unconditionally return true, as simple classes are unsubscripted and
        # could thus be said to only have ignorable arguments. Look. Semantics.
        return True

    # ..................{ PRIVATE ~ methods                  }..................
    def _is_subhint_branch(self, branch: TypeHint) -> bool:
        # print(f'Entering ClassTypeHint._is_subhint_branch({self}, {branch})...')
        # print(f'{branch}._is_args_ignorable: {branch._is_args_ignorable}')
        # print(f'{self}._origin: {self._origin}')
        # print(f'{branch}._origin: {branch._origin}')

        # print(f'{self}._hint: {self._hint}')
        # print(f'{branch}._hint: {branch._hint}')
        # # print(f'{repr(self)}._origin.__args__: {self._origin.__args__}')
        # print(f'{repr(self)}._origin.__parameters__: {self._origin.__parameters__}')
        # # print(f'{repr(branch)}._origin.__args__: {branch._origin.__args__}')
        # print(f'{repr(branch)}._origin.__parameters__: {branch._origin.__parameters__}')
        # print(f'{repr(self)}._is_args_ignorable: {self._is_args_ignorable}')

        #FIXME: Actually, let's avoid the implicit numeric tower for now.
        #Explicit is better than implicit and we really strongly disagree with
        #this subsection of PEP 484, which does more real-world harm than good.
        # # Numeric tower:
        # # https://peps.python.org/pep-0484/#the-numeric-tower
        # if self._origin is float and branch._origin in {float, int}:
        #     return True
        # if self._origin is complex and branch._origin in {complex, float, int}:
        #     return True

        # Return true only if...
        return (
            # That class is unsubscripted (and thus *NOT* a subscripted generic)
            # *AND*...
            branch._is_args_ignorable and
            # The unsubscripted type originating this class is a subclass of the
            # unsubscripted type originating that class.
            issubclass(self._origin, branch._origin)
        )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide :pep:`649` **data submodule.**

This submodule exercises :pep:`649` support for unquoted forward references
inside type hinting contexts implemented in the :func:`beartype.beartype`
decorator by declaring unit tests exercising these references. For safety, these
tests are intentionally isolated from the main test suite.

Caveats
-------
**This submodule requires the active Python interpreter to target at least
Python 3.14.0.** If this is *not* the case, importing this submodule raises
:exc:`NameError` exceptions.
'''

# ....................{ TESTS ~ getter                     }....................
def unit_test_get_pep649_hintable_annotations() -> None:
    '''
    Test the :pep:`649`-compliant implementation of the private
    :mod:`beartype._util.hint.pep.proposal.pep649.get_pep649_hintable_annotations`
    getter under Python >= 3.14.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar import BeartypeDecorHintPep649Exception
    # from beartype.typing import (
    #     Generic,
    #     Tuple,
    #     TypeVarTuple,
    #     Unpack,
    # )
    from beartype._util.hint.pep.proposal.pep649 import (
        get_pep649_hintable_annotations)
    from pytest import raises

    # # ....................{ CALLABLES                      }....................
    # @beartype
    # def of_all_the_grace(
    #     and_to_the_damp_leaves: ItsMotions,
    #     *and_beauty_that_endued: *Ts
    # # ) -> Tuple[ItsMotions, *Ts]:
    # ) -> object:
    #     '''
    #     Arbitrary callable simply returning the tuple of all passed variadic
    #     positional parameters prepended by the passed generic, decorated by
    #     :func:`beartype.beartype` and correctly annotated by:
    #
    #     * A variadic positional parameter typed as a :pep:`649`-compliant
    #       unpacked type variable tuple hint.
    #     * A return typed as a :pep:`484`-compliant tuple type hint subscripted
    #       by the same :pep:`649`-compliant unpacked type variable tuple hint.
    #     '''
    #
    #     # Return this variadic positional parameter prepended by this generic.
    #     return (and_to_the_damp_leaves,) + and_beauty_that_endued
    #
    # # ....................{ PASS                           }....................
    # # Assert that this callable returns a tuple of all passed positional
    # # parameters.
    # assert of_all_the_grace(
    #     render_up_its_majesty, 'Of all the grace', 'and beauty', 'that endued') == (
    #     render_up_its_majesty, 'Of all the grace', 'and beauty', 'that endued')
    #
    # # ....................{ FAIL                           }....................
    # # Assert that @beartype raises the expected exception when decorating a
    # # callable annotated by a hint unpacking a PEP 649-noncompliant object
    # # (i.e., *ANY* object other than a PEP 649-compliant type variable tuple).
    # with raises(BeartypeDecorHintPep649Exception):
    #     @beartype
    #     def scatter_its_music(
    #         *on_the_unfeeling_storm: Unpack[Tuple[str]]) -> str:
    #         return on_the_unfeeling_storm[0]



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **PEP-noncompliant NumPy type hint reducers** (i.e., low-level
callables converting higher-level type hints defined by the third-party
:mod:`numpy` package that do *not* comply with any specific PEP but are
nonetheless shallowly supported by :mod:`beartype` to lower-level type hints
more readily consumable by :mod:`beartype`).

This private submodule is *not* intended for importation by downstream callers.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# CAUTION: The top-level of this module should avoid importing from third-party
# optional libraries, both because those libraries cannot be guaranteed to be
# either installed or importable here *AND* because those imports are likely to
# be computationally expensive, particularly for imports transitively importing
# C extensions (e.g., anything from NumPy or SciPy).
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from beartype.roar import BeartypeDecorHintNonpepNumpyException
from beartype.typing import (
    Annotated,
    Any,
)
from beartype._data.typing.datatypingport import Hint
from beartype._util.api.external.utilnumpy import (
    get_numpy_dtype_type_abcs,
    make_numpy_dtype,
)
from beartype._util.cache.utilcachecall import callable_cached
from beartype._util.hint.pep.utilpepget import get_hint_pep_args
from beartype._util.utilobject import is_object_hashable

# ....................{ REDUCERS                           }....................
@callable_cached
def reduce_hint_numpy_ndarray(hint: Hint, exception_prefix: str) -> Hint:
    '''
    Reduce the passed **PEP-noncompliant typed NumPy array** (i.e.,
    subscription of the third-party :attr:`numpy.typing.NDArray` type hint
    factory) to the equivalent PEP-compliant beartype validator validating
    arbitrary objects be instances of that array type -- which has the
    substantial merit of already being well-supported, well-tested, and
    well-known to generate optimally efficient type-checking by the
    :func:`beartype.beartype` decorator.

    Technically, beartype could instead explicitly handle typed NumPy arrays
    throughout the codebase. Of course, doing so would yield *no* tangible
    benefits while imposing a considerable maintenance burden.

    This reducer is memoized for efficiency.

    Parameters
    ----------
    hint : Hint
        PEP-noncompliant typed NumPy array to be reduced.
    exception_prefix : str
        Human-readable substring prefixing raised exception messages.

    Returns
    -------
    Hint
        This PEP-noncompliant typed NumPy array reduced to a PEP-compliant type
        hint supported by :mod:`beartype`.

    Raises
    ------
    BeartypeDecorHintNonpepNumpyException
        If this hint is a typed NumPy array but either:

        * *Not* subscripted by exactly two arguments.
        * Subscripted by exactly two arguments but whose second argument is
          neither:

          * A **NumPy data type** (i.e., :class:`numpy.dtype` instance).
          * An object coercible into a NumPy data type by passing to the
            :meth:`numpy.dtype.__init__` method.
    '''

    # ..................{ IMPORTS                            }..................
    # Defer heavyweight imports until *AFTER* validating this hint to be a typed
    # NumPy array. Why? Because these imports are *ONLY* safely importable if
    # this hint is a typed NumPy array. Why? Because instantiating this hint
    # required these imports. QED.
    #
    # Note that third-party packages should typically *ONLY* be imported via
    # utility functions raising human-readable exceptions when those packages
    # are either uninstalled or unimportable. In this case, however, NumPy will
    # almost *ALWAYS* be importable. Why? Because this hint was externally
    # instantiated by the user by first importing the "numpy.typing.NDArray"
    # attribute passed to this getter.
    from beartype.vale import (
        IsAttr,
        IsEqual,
        IsSubclass,
    )
    from numpy import ndarray  # pyright: ignore
    from numpy.typing import NDArray  # type: ignore[attr-defined]

    #FIXME: Consider submitting an upstream issue about this. We don't
    #particularly feel like arguing tonight, because that's a lonely hill.

    # If this hint is the unsubscripted "NDArray" type hint, this hint
    # permissively matches *ALL* NumPy arrays rather than strictly matching
    # *ONLY* appropriately typed NumPy arrays. In this case, reduce this hint
    # to the untyped "numpy.ndarray" class.
    #
    # Note the similar test matching the subscripted "NDArray[Any]" hint below.
    # Moreover, note this test *CANNOT* be performed elsewhere (e.g., by
    # adding "HintSignNumpyArray" to the "HINT_SIGNS_ORIGIN_ISINSTANCEABLE"
    # frozen set of all signs whose unsubscripted type hint factories are
    # shallowly type-checkable). Why? Because the "NDArray" type hint factory
    # violates type hinting standards. Specifically, this factory implicitly
    # subscripts *AND* parametrizes itself with the "numpy.ScalarType" type
    # variable bounded above by the "numpy.generic" abstract base class for
    # NumPy scalars.
    #
    # We have *NO* idea why NumPy does this. This implicit behaviour is
    # semantically lossy rather than lossless and thus arguably constitutes an
    # upstream bug. Why? Because this behaviour violates:
    # * The NumPy API. The "NDArray" type hint factory is subscriptable by more
    #   than merely NumPy scalar types. Ergo, "NDArray" is semantically
    #   inaccurate!
    # * PEP 484, which explicitly standardizes an equivalence between
    #   unsubscripted type hint factories and the same factories subscripted by
    #   the "typing.Any" singleton. However, "NDArray" is *MUCH* semantically
    #   narrower than and thus *NOT* equivalent to "NDArray[Any]"!
    #
    # Of course, upstream is unlikely to see it that way. We're *NOT* dying on
    # an argumentative hill about semantics. Upstream makes the rules. Do it.
    if hint is NDArray:
        return ndarray
    # Else, this hint is *NOT* the unsubscripted "NDArray" type hint.

    # ..................{ CONSTANTS                          }..................
    # Frozen set of all NumPy scalar data type abstract base classes (ABCs).
    NUMPY_DTYPE_TYPE_ABCS = get_numpy_dtype_type_abcs()

    # ..................{ ARGS                               }..................
    # Objects subscripting this hint if any *OR* the empty tuple otherwise.
    hint_args = get_hint_pep_args(hint)

    # If this hint was *NOT* subscripted by exactly two arguments, this hint is
    # malformed as a typed NumPy array. In this case, raise an exception.
    if len(hint_args) != 2:
        raise BeartypeDecorHintNonpepNumpyException(
            f'{exception_prefix}typed NumPy array {repr(hint)} '
            f'not subscripted by exactly two arguments.'
        )
    # Else, this hint was subscripted by exactly two arguments.

    # Data type subhint subscripting this hint. Yes, the "numpy.typing.NDArray"
    # type hint bizarrely encapsulates its data type argument into a private
    # "numpy._DTypeMeta" type subhint. Why? We have absolutely no idea, but we
    # have no say in the matter. NumPy, you're on notice for stupidity.
    hint_dtype_subhint = hint_args[1]

    # Objects subscripting this subhint if any *OR* the empty tuple otherwise.
    hint_dtype_subhint_args = get_hint_pep_args(hint_dtype_subhint)

    # If this hint was *NOT* subscripted by exactly one argument, this subhint
    # is malformed as a data type subhint. In this case, raise an exception.
    if len(hint_dtype_subhint_args) != 1:
        raise BeartypeDecorHintNonpepNumpyException(
            f'{exception_prefix}typed NumPy array {repr(hint)} '
            f'data type subhint {repr(hint_dtype_subhint)} '
            f'not subscripted by exactly one argument.'
        )
    # Else, this subhint was subscripted by exactly one argument.

    # Data type-like object subscripting this subhint. Look, just do it.
    hint_dtype_like = hint_dtype_subhint_args[0]

    # If this dtype-like is "typing.Any", this hint permissively matches *ALL*
    # NumPy arrays rather than strictly matching *ONLY* appropriately typed
    # NumPy arrays. In this case, reduce this hint to the untyped
    # "numpy.ndarray" class.
    #
    # Note the similar test matching the unsubscripted "NDArray" hint above.
    if hint_dtype_like is Any:
        return ndarray  # pyright: ignore
    # Else, this dtype-like is *NOT* "typing.Any".

    # ..................{ REDUCTION                          }..................
    # Equivalent nested beartype validator reduced from this hint.
    hint_validator = None  # type: ignore[assignment]

    # If...
    if (
        # This dtype-like is hashable *AND*...
        is_object_hashable(hint_dtype_like) and
        # This dtype-like is a scalar data type abstract base class (ABC)...
        hint_dtype_like in NUMPY_DTYPE_TYPE_ABCS
    ):
        # Then avoid attempting to coerce this possibly non-dtype into a proper
        # dtype. Although NumPy previously silently coerced these ABCs into
        # dtypes (e.g., from "numpy.floating" to "numpy.float64"), recent
        # versions of NumPy now emit non-fatal deprecation warnings on doing so
        # and will presumably raise fatal exceptions in the near future:
        #     >>> import numpy as np
        #     >>> np.dtype(np.floating)
        #     DeprecationWarning: Converting `np.inexact` or `np.floating` to a
        #     dtype is deprecated. The current result is `float64` which is not
        #     strictly correct.
        #
        # Instead, we follow mypy's lead. Presumably defined somewhere in the
        # incredibly complex innards of NumPy's mypy plugin (which we admittedly
        # failed to grep despite ~~wasting~~ "investing" several hours in doing
        # so), mypy treats subscriptions of the "numpy.typing.NDArray" type hint
        # factory by one of these ABCs (rather than either a scalar or proper
        # dtype) as a type inheritance (rather than object equality) relation.
        # Since this is sensible, we do too.

        # Equivalent nested beartype validator reduced from this hint.
        hint_validator = (
            IsAttr['dtype', IsAttr['type', IsSubclass[hint_dtype_like]]])
    # Else, this dtype-like is either unhashable *OR* not such an ABC. In this
    # case...
    else:
        # Proper dtype coerced from this possibly non-dtype.
        hint_dtype = make_numpy_dtype(
            dtype=hint_dtype_like,
            exception_prefix=exception_prefix,
            exception_cls=BeartypeDecorHintNonpepNumpyException,
        )

        # Equivalent nested beartype validator reduced from this hint.
        hint_validator = IsAttr['dtype', IsEqual[hint_dtype]]

    # Replace the usually less readable representation of this validator with
    # the usually more readable representation of this hint (e.g.,
    # "numpy.ndarray[typing.Any, numpy.float64]").
    hint_validator.get_repr = repr(hint)

    # Return this validator annotating the NumPy array type.
    return Annotated[ndarray, hint_validator]  # type: ignore[return-value]  # pyright: ignore



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
Project-wide **Equinox integration tests.**

This submodule functionally tests the :mod:`beartype` package against the
third-party :mod:`equinox` package.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid inscrutable issues with previously run unit tests requiring
# forked subprocess isolation, avoid attempting to conditionally skip
# JAX-dependent integration tests with standard mark decorators: e.g.,
#     @skip_unless_package('jax')  # <-- *NEVER DO THIS* srsly. never.
#     @skip_unless_package('jaxtyping')  # <-- *NEVER DO THIS EITHER* bad is bad
#
# Why? Because even the mere act of attempting to decide whether JAX is
# importable at early test collection time causes the first unit test isolated
# to a forked subprocess to raise the following suspicious exception:
# E       pytest.PytestUnraisableExceptionWarning: Exception ignored in:
#         <function _at_fork at 0x7f3865738310>
# E       Traceback (most recent call last):
# E         File "/home/leycec/py/conda/envs/ionyou_dev/lib/python3.11/site-packages/jax/_src/xla_bridge.py", line 112, in _at_fork
# E           warnings.warn(
# E       RuntimeWarning: os.fork() was called. os.fork() is incompatible
#         with multithreaded code, and JAX is multithreaded, so this will
#         likely lead to a deadlock.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
import pytest
from beartype_test._util.mark.pytskip import skip

# ....................{ TESTS                              }....................
# Note that any importation whatsoever from JAX is dangerous and *MUST* be
# isolated to a JAX-specific subprocess. Failure to do so raises the above
# warning. Honestly, what a pain. Come on, JAX. Work with us here. *sigh*
@pytest.mark.run_in_subprocess
def test_equinox_filter_jit() -> None:
    '''
    Functional test validating that the :mod:`beartype` package successfully
    type-checks callables decorated by the third-party
    :func:`equinox.filter_jit` decorator annotated by type hints by the
    third-party :mod:`jaxtyping` package.

    See Also
    --------
    https://github.com/beartype/beartype/issues/368
        Issue strongly inspiring this functional test.
    '''

    # ....................{ IMPORTS ~ early                }....................
    # Defer test-specific imports.
    #
    # Note that JAX-dependent packages are *NOT* safely importable from here.
    from beartype import beartype
    from beartype.roar import BeartypeCallHintParamViolation
    from beartype._util.module.utilmodtest import is_package
    from pytest import raises

    # If any requisite JAX package is unimportable, silently reduce to a noop.
    #
    # Note that merely testing the importability of a JAX package emits warnings
    # in unpredictably hardware-dependent edge cases. Since that then induces
    # test failure, these tests necessarily ignore these warnings. For example,
    # if the low-level C-based "jaxlib" package was compiled on a newer system
    # supporting the assembly-level AVX instruction set that the current system
    # fails to support, these tests would emit this warning:
    #     E   RuntimeError: This version of jaxlib was built using AVX
    #         instructions, which your CPU and/or operating system do not
    #         support. You may be able work around this issue by building jaxlib
    #         from source.
    if not (
        is_package('equinox', is_warnings_ignore=True) and
        is_package('jax', is_warnings_ignore=True) and
        is_package('jaxtyping', is_warnings_ignore=True)
    ):
        return
    # Else, all requisite JAX packages is importable.

    # ....................{ IMPORTS ~ late                 }....................
    # Defer JAX-dependent imports.
    from equinox import filter_jit
    from jax import numpy as jax_numpy
    from jaxtyping import (
        Array,
        Float,
    )

    # ....................{ LOCALS                         }....................
    # Type hint matching a JAX array of floating-point numbers.
    JaxArrayOfFloats = Float[Array, '']

    # JAX array of arbitrary floating-point numbers.
    of_those_beloved_eyes = jax_numpy.array(1.0)

    # ....................{ CALLABLES                      }....................
    @beartype
    @filter_jit
    def as_if_their_genii(
        were_the_ministers: JaxArrayOfFloats) -> JaxArrayOfFloats:
        '''
        Arbitrary callable decorated first by :func:`equinox.filter_jit` and
        then by :func:`beartype.beartype`, exercising a well-known edge case.
        '''

        # Do it because it feels good, one-liner.
        return were_the_ministers + 1


    @beartype
    @filter_jit
    def appointed_to_conduct_him(
        to_the_light: JaxArrayOfFloats) -> JaxArrayOfFloats:
        '''
        Arbitrary callable decorated first by :func:`equinox.filter_jit` and
        then by :func:`beartype.beartype`, exercising a well-known edge case.

        Note that fully exercising this edge case requires defining not merely
        one but *TWO* callables decorated in this order. Before this issue was
        resolved, :func:`beartype.beartype` *literally* swapped the code objects
        of these two callables. Believe it or not, it's beartype! O_o
        '''

        # Do it because it feels right, one-liner.
        return to_the_light - 1

    # ....................{ PASS                           }....................
    # Assert that these callables return the expected values when passed valid
    # parameters satisfying the type hints annotating these callables.
    assert as_if_their_genii(of_those_beloved_eyes) == (
        jax_numpy.array(2.0))
    assert appointed_to_conduct_him(of_those_beloved_eyes) == (
        jax_numpy.array(0.0))

    # ....................{ FAIL                           }....................
    # Assert that these callables raise the expected exceptions when passed
    # invalid parameters violating the type hints annotating these callables.
    with raises(BeartypeCallHintParamViolation):
        as_if_their_genii('Appointed to conduct him to the light')
    with raises(BeartypeCallHintParamViolation):
        appointed_to_conduct_him('Of those belovèd eyes, the Poet sate')


#FIXME: Resurrect this test *AFTER* resurrecting Equinox integration, please.
@skip('Currently broken due to @beartype temporarily dropping Equinox support.')
def test_equinox_module_subclass() -> None:
    '''
    Functional test validating that the :mod:`beartype` package successfully
    type-checks subclasses of superclasses defined by the third-party Equinox
    package annotated by type hints by the third-party :mod:`jaxtyping` package.

    See Also
    --------
    https://github.com/patrick-kidger/equinox/issues/584
        Upstream Equinox issue strongly inspiring this functional test.
    '''

    # ....................{ IMPORTS ~ early                }....................
    # Defer test-specific imports.
    #
    # Note that JAX-dependent packages are *NOT* safely importable from here.
    from beartype import beartype
    from beartype.roar import BeartypeCallHintParamViolation
    from beartype._util.module.utilmodtest import is_package
    from pytest import raises

    #FIXME: *EVEN THIS ISN"T SAFE.* Any importation whatsoever from JAX is
    #dangerous and *MUST* be isolated to a subprocess. Honestly, what a pain.
    #See similar logic in "test_jax" also requiring a similar resolution.
    # If any requisite JAX package is unimportable, silently reduce to a noop.
    #
    # Note that merely testing the importability of a JAX package emits warnings
    # in unpredictably hardware-dependent edge cases. Since that then induces
    # test failure, these tests necessarily ignore these warnings. For example,
    # if the low-level C-based "jaxlib" package was compiled on a newer system
    # supporting the assembly-level AVX instruction set that the current system
    # fails to support, these tests would emit this warning:
    #     E   RuntimeError: This version of jaxlib was built using AVX
    #         instructions, which your CPU and/or operating system do not
    #         support. You may be able work around this issue by building jaxlib
    #         from source.
    if not (
        is_package('equinox', is_warnings_ignore=True) and
        is_package('jax', is_warnings_ignore=True) and
        is_package('jaxtyping', is_warnings_ignore=True)
    ):
        return
    # Else, all requisite JAX packages is importable.

    # ....................{ IMPORTS ~ late                 }....................
    # Defer JAX-dependent imports.
    from equinox import Module
    from jax import numpy as jax_numpy
    from jaxtyping import (
        Array,
        Float,
    )

    # ....................{ CLASSES                        }....................
    @beartype
    class EquinoxModule(Module):
        '''
        Arbitrary subclass of the :class:`equinox.Module` superclass decorated
        by the :func:`beartype.beartype` decorator.
        '''

        # ....................{ CLASS VARS                 }....................
        float_array: Float[Array, '']
        '''
        Arbitrary class variable annotated by a :mod:`jaxtyping` type hint.
        '''

        # ....................{ METHODS                    }....................
        def munge_array(self, python_bool: bool) -> Float[Array, '']:
            '''
            Arbitrary method accepting a trivial object to be type-checked.
            '''

            # Arbitrary one-liner is arbitrary.
            return self.float_array + 1

    # ....................{ LOCALS                         }....................
    # JAX-based NumPy array containing arbitrary data.
    jax_array = jax_numpy.array(1.0)

    # Arbitrary instance of the above Equinox subclass.
    equinox_module = EquinoxModule(jax_array)

    # ....................{ FAIL                           }....................
    # Assert that this @beartype-decorated method of this Equinox instance
    # raises the expected type-checking violation exception when passed an
    # invalid parameter violating the type hint annotating that parameter.
    with raises(BeartypeCallHintParamViolation):
        equinox_module.munge_array('A string is not a boolean.')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright (c) 2014-2025 Beartype authors.
# See "LICENSE" for further details.

'''
**Callable parameter getter utility unit tests.**

This submodule unit tests the public API of the private
:mod:`beartype._util.func.arg.utilfuncargget` submodule.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable test errors, avoid importing from
# package-specific submodules at module scope.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

# ....................{ TESTS ~ name                       }....................
def test_get_func_arg_name_first_or_none() -> None:
    '''
    Test the
    :func:`beartype._util.func.arg.utilfuncargget.get_func_arg_name_first_or_none`
    getter.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar._roarexc import _BeartypeUtilCallableException
    from beartype._util.func.arg.utilfuncargget import (
        get_func_arg_name_first_or_none)
    from beartype_test.a00_unit.data.func.data_func import (
        func_args_0_wrapped,
        func_args_1_varpos_wrapped,
        func_args_3_flex_mandatory_optional_varkw_wrapped,
    )
    from pytest import raises

    # ....................{ PASS                           }....................
    # Assert this getter returns "None" for argumentless callables.
    assert get_func_arg_name_first_or_none(func_args_0_wrapped) is None

    # Assert this getter returns the expected names of the first parameters
    # accepted by argumentative callables.
    assert get_func_arg_name_first_or_none(func_args_1_varpos_wrapped) == (
        'and_in_her_his_one_delight')
    assert get_func_arg_name_first_or_none(
        func_args_3_flex_mandatory_optional_varkw_wrapped) == (
        'and_the_wolf_tracks_her_there')

    # ....................{ FAIL                           }....................
    # Assert this getter raises the expected exception when passed a C-based
    # callable.
    with raises(_BeartypeUtilCallableException):
        get_func_arg_name_first_or_none(iter)


def test_get_func_arg_names() -> None:
    '''
    Test the
    :func:`beartype._util.func.arg.utilfuncargget.get_func_arg_names`
    getter.
    '''

    # ....................{ IMPORTS                        }....................
    # Defer test-specific imports.
    from beartype.roar._roarexc import _BeartypeUtilCallableException
    from beartype._util.func.arg.utilfuncargget import get_func_arg_names
    from beartype_test.a00_unit.data.func.data_func import (
        func_args_0_wrapped,
        func_args_1_varpos_wrapped,
        func_args_3_flex_mandatory_optional_varkw_wrapped,
        func_args_5_flex_mandatory_varpos_kwonly_varkw_wrapped,
    )
    from beartype_test.a00_unit.data.func.data_pep570 import (
        func_args_10_all_except_flex_mandatory)
    from pytest import raises

    # ....................{ PASS                           }....................
    # Assert this getter when passed an argumentless callable returns the empty
    # list.
    assert get_func_arg_names(func_args_0_wrapped) == ()

    # Assert this getter when passed argumentative callables returns the
    # expected lists of the names of the parameters accepted by these callables.
    assert get_func_arg_names(func_args_1_varpos_wrapped) == (
        'and_in_her_his_one_delight',)
    assert get_func_arg_names(
        func_args_3_flex_mandatory_optional_varkw_wrapped) == (
        'and_the_wolf_tracks_her_there',
        'how_hideously',
        'rude_bare_and_high',
    )
    assert get_func_arg_names(
        func_args_5_flex_mandatory_varpos_kwonly_varkw_wrapped) == (
        'we_are_selfish_men',
        'oh_raise_us_up',
        'return_to_us_again',
        'and_give_us',
        'manners_virtue_freedom_power',
    )

    assert get_func_arg_names(
        func_args_10_all_except_flex_mandatory) == (
        'in_solitude_i_wander',
        'through_the_vast_enchanted_forest',
        'the_surrounding_skies',
        'torn_apart_by',
        'rain_is_pouring_down',
        'the_darkened_oaks_are_my_only_shelter',
        'red_leaves_are_blown_by',
        'an_ebony_raven_now_catches',
        'in_the_rain_my_tears_are_forever_lost',
        'sitting_in_calmness',
    )

    # ....................{ FAIL                           }....................
    # Assert this getter raises the expected exception when passed a C-based
    # callable.
    with raises(_BeartypeUtilCallableException):
        get_func_arg_names(iter)
